/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 1.1.0
          File Name: imagefilefactory.cpp
         Start Date: 25/02/2010
       Last Changed: $LastChangedDate: 2011-06-12 16:41:13 +1200 (Sun, 12 Jun 2011) $
           Revision: $Revision: 310 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
   File Description: ImageFileFactory Class implementation
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "imagefilefactory.h"

ImageFileFactory::ImageFileFactory(QObject *parent) :
    QObject(parent)
{
    rawPlugin = 0;
}

ImageFileFactory::~ImageFileFactory() {

}

int ImageFileFactory::loadImage(QString filename,
                                ImageFile** imageFile) {

    qDebug() << "Attempting to load file " << filename;

    /* Ask each plugin to see if we can find one that can open the specified
       file.
     */
    foreach (ImageFilePlugin *ifp, plugins) {
        qDebug() << "Trying File Plugin " << ifp->getName();
        if (ifp->fileIsSupported(filename) &&
            ifp->getFeatures().testFlag(ImageFilePlugin::F_Read)) {
            qDebug() << "Plugin reports file is supported. Attempting to load.";
            // The file is supported! Go open it
            *imageFile = ifp->createImageFile(filename);

            if (*imageFile == 0) {
                qDebug() << "Load attempt failed. Continuing search for suitable"
                        " plugin.";
            } else {
                qDebug() << "Load successfull. Returning.";
                return errors::E_SUCCESS;
            }
        }
    }

    /* If we got this far then none of the file format plugins could understand
       the file we specified. If we have a Raw file format plugin then use that */
    if (rawPlugin != 0) {
        qDebug() << "Could not find suitable standard plugin to load file. "
                    "Attempting raw file handler...";
        *imageFile = rawPlugin->createImageFile(filename);
        if (*imageFile == 0) {
            qDebug() << "Load attempt failed. NULL was returned. Assuming unknown file format.";
            return errors::E_UNKNOWN_FILE_FORMAT;
        }
        return errors::E_SUCCESS;
    }

    qDebug() << "No loaded plugin was able to handle file. Aborting.";

    /* If we got this far then we don't have anything that can load the file.
       Give up. */

    return errors::E_UNKNOWN_FILE_FORMAT;
}

int ImageFileFactory::createImage(QString filename,
                                  ImageFile** imageFile,
                                  long long expectedSize) {
    return createImage(filename,
                       imageFile,
                       expectedSize,
                       ImageFileFactory::F_NoHints);
}

int ImageFileFactory::createImage(QString filename,
                                  ImageFile** imageFile,
                                  long long expectedSize,
                                  FileHints hints) {

    // Setup the image file plugins load hints
    ImageFilePlugin::LoadHints loadHints;
    if (hints.testFlag(F_DontCache)) loadHints |= ImageFilePlugin::LH_NoCache;

    // Because we want to create the image.
    loadHints |= ImageFilePlugin::LH_Create;

    qDebug() << "Trying to find plugin with which to create image file...";
    foreach(ImageFilePlugin *ifp, plugins) {
        qDebug() << "Trying plugin " << ifp->getName();
        if (ifp->fileIsSupported(filename) &&
            ifp->getFeatures().testFlag(ImageFilePlugin::F_Write)) {
            qDebug() << "Plugin reports file is supported. Attempting to create.";
            *imageFile = ifp->createImageFile(filename,loadHints,expectedSize);

            if (imageFile == 0) {
                qDebug() << "Create attempt failed. Continuing search for suitable "
                        "plugin.";
            } else {
                qDebug() << "Create successfull. Returning.";
                return errors::E_SUCCESS;
            }
        }
    }

    // If we got this far then we couldnt find a plugin above to open it. Try
    // the raw plugin (if we have one)
    if (rawPlugin != 0) {
        qDebug() << "Could not find suitable standard plugin to load file. "
                    "Attempting raw file handler...";
        if (rawPlugin->fileIsSupported(filename)) {
            qDebug() << "Plugin reports format is supported. Trying to load...";
            *imageFile = rawPlugin->createImageFile(filename,loadHints,expectedSize);
            if (imageFile == 0) {
                qDebug() << "Load attempt failed. NULL was returned.";
                return errors::E_LOAD_FAILED;
            }
            return errors::E_SUCCESS;
        }
    }

    return errors::E_UNKNOWN_FILE_FORMAT;
}

QString ImageFileFactory::getExtensionsFromFormatString(QString formatString) {
    int openBraceLoc = formatString.indexOf("(",0);

    QString temp = formatString.mid(openBraceLoc+1);
    int closeBraceLoc = temp.length()+ temp.lastIndexOf(")",0);
    temp = temp.left(closeBraceLoc);
    return temp;
}

QStringList ImageFileFactory::getFileNameFilters(bool includeAllFilter) {
    qDebug() << "Getting filename filters...";
    QStringList formatStrings;

    QString asfString = "All Supported Images (";

    foreach(ImageFilePlugin *plugin,plugins) {
        QStringList nameFilters = plugin->getNameFilters();
        formatStrings.append(nameFilters);

        foreach(QString filter,nameFilters) {
            if (!asfString.endsWith("("))
                if (!asfString.endsWith(" "))
                    asfString.append(" ");

            asfString.append(getExtensionsFromFormatString(filter));
        }
    }

    if (rawPlugin) {
        QStringList nameFilters = rawPlugin->getNameFilters();
        foreach(QString filter,nameFilters) {
            if (!asfString.endsWith("("))
                if (!asfString.endsWith(" "))
                    asfString.append(" ");

            asfString.append(getExtensionsFromFormatString(filter));
        }
        formatStrings.append(nameFilters);
    }

    asfString.append(")");

    // Only include the "All SUpported Images" filter if it was requested.
    if (includeAllFilter)
        formatStrings.prepend(asfString);

    return formatStrings;
}

int ImageFileFactory::loadPlugins(QString pluginDirectory) {

    qDebug() << "Image Format plugin dir: " << pluginDirectory;

    // If there are plugins loaded, unload them.
    while (!plugins.isEmpty())
        delete plugins.takeFirst();

    QDir pluginsDir = QDir(pluginDirectory);

    foreach(QString fileName, pluginsDir.entryList(QDir::Files)) {
        qDebug() << "Attempting to load as format plugin file: " << fileName;
        QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
        QObject *plugin = loader.instance();
        if (plugin) {
            ImageFilePlugin *ifp = qobject_cast<ImageFilePlugin *>(plugin);
            if (ifp) {
                qDebug() << "Successfully loaded " << fileName;
                /* Raw plugins are treated differently as they will accept anything
                   (there isn't any way for a raw plugin to tell if it is a valid
                    image or not). As such they need to be tried last when
                    attempting to open an image.
                 */
                if (ifp->getFeatures().testFlag(ImageFilePlugin::F_RawFormat)) {
                    // If we already have a raw plugin then unload it.
                    if (rawPlugin != 0) delete rawPlugin;
                    rawPlugin = ifp;
                } else {
                    plugins.append(ifp);
                }
            } else {
                qDebug() << "Failed to load " << fileName;
            }
        }
    }

    if (plugins.isEmpty()) return errors::E_NO_FORMAT_PLUGINS;
    return errors::E_SUCCESS;
}
