/*
-----------------------------------------------------
   Application Name: "DIMG"
Application Version: 1.1.0
          File Name: imagefileutils.cpp
         Start Date: 27/02/2010
       Last Changed: $LastChangedDate: 2011-06-12 17:44:18 +1200 (Sun, 12 Jun 2011) $
           Revision: $Revision: 314 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
   File Description: ImageFileUtils Class implementation
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "imagefileutils.h"

#include "commentprompt.h"

#include <QApplication>

void ImageFileUtils::setSettings(QSettings *settings) {
    this->settings = settings;
}

void ImageFileUtils::initialise() {
    qDebug() << "ImageFileUtils initialising...";

    Q_ASSERT_X(!settings.isNull(),
               "ImageFileUtils::initialise()",
               "Settings object has not been set.");

    factory.loadPlugins(settings->value("/Plugins/ImageFormatDir",
                                        QApplication::applicationDirPath()).toString());
}

int ImageFileUtils::openImage(ImageFile ** imageFile) {
    qDebug() << "ImageFileUtils::openImage()";

    QString filter = factory.getFileNameFilters().join(";;");

    qDebug() << "Filename filter: " << filter;

    QString lastOpenDir = settings->value("/State/Dirs/LastOpenDir/","./").toString();

    QString filename = QFileDialog::getOpenFileName(0,
                                                    "Open Disk Image...",
                                                    lastOpenDir,
                                                    filter);

    int result = openImage(imageFile, filename);

    if (result == errors::E_USER_CANCELED) return result;

    if (imageFile == 0) {
        qDebug() << "Hmmn. We seem to have got null back.";
        return result;
    }

    if (result != errors::E_SUCCESS) {
        qDebug() << "Failed to open image.";
        return result;
    }

    settings->setValue("/State/Dirs/LastOpenDir/",QFileInfo(filename).absolutePath());

    qDebug() << "Image opened successfully.";

    return result;
}

int ImageFileUtils::openImage(ImageFile **imageFile, QString filename) {
    QFile fi(filename);

    if (!fi.exists()) {
        qDebug() << "Requested file does not exist.";
        imageFile = 0;
        return errors::E_USER_CANCELED;
    }

    int result = factory.loadImage(filename,imageFile);


    if (imageFile == 0) {
        qDebug() << "Hmmn. We seem to have got null back.";
        return result;
    }

    return result;
}

QString ImageFileUtils::getSaveAsFilename(QString filename) {
    qDebug() << "ImageFileUtils::getSaveAsFilename()";

    QStringList filterStrings = factory.getFileNameFilters(false);
    QString filter = filterStrings.join(";;");
    QString lastSaveDir = "";
    QString targetFilter = "";

    if (filename.isEmpty())
        lastSaveDir = settings->value("/State/Dirs/LastSaveDir/",
                           settings->value("/State/Dirs/LastOpenDir/","./")).toString();
    else {
        lastSaveDir = filename;

        // Now we try to select a filter that matches the current filename so
        // that the user won't accidentally convert the image to another
        // format. This seems needlessly complex btw.
        QString suffix = QFileInfo(filename).completeSuffix();
        if (!suffix.startsWith(".")) suffix.prepend(".");
        QString regexpString = suffix.replace(".","\\.");
        QRegExp regexp(regexpString);
        regexp.setCaseSensitivity(Qt::CaseInsensitive);
        QStringList availableFilterStrings = filterStrings.filter(regexp);
        if (!availableFilterStrings.isEmpty()) {
            QString currentFilterString = availableFilterStrings.first();
            qDebug() << "Target filter string: " << currentFilterString;
            targetFilter = currentFilterString;
        }
    }

    QString saveFilename = QFileDialog::getSaveFileName(0,
                                                    "Save Disk Image As...",
                                                    lastSaveDir,
                                                    filter,
                                                    &targetFilter);

    if (settings->contains("/State/Dirs/LastSaveDir/"))
        settings->setValue("/State/Dirs/LastSaveDir/",QFileInfo(saveFilename).absolutePath());
    else
        settings->setValue("/State/Dirs/LastOpenDir/",QFileInfo(saveFilename).absolutePath());

    return saveFilename;
}

int ImageFileUtils::createImage(ImageFile ** imageFile, long long expectedSize) {
    qDebug() << "ImageFileUtils::createImage()";

    QString filename = getSaveAsFilename("");

    /*
     * We will create the image file with caching disabled because:
     *  1) If the image is cacheable DIMG will have cached it on its own
     *  2) If the image is uncacheable there is no point turning on caching.
     */
    return factory.createImage(filename,
                               imageFile,
                               expectedSize,
                               ImageFileFactory::F_DontCache);
}


int ImageFileUtils::saveImageAs(ImageFile **oldImageFile, ImageFile **newImageFile) {
    qDebug() << "ImageFileUtils::saveImageAs()";
    if (oldImageFile == 0) {
        qDebug() << "ERROR: No old image file passed in.";
        return errors::E_BAD_POINTER;
    }
    if ((*newImageFile) != 0) {
        qDebug() << "ERROR: newImageFile is not null. Overwriting may result in "
                "memory leaks.";
        return errors::E_BAD_POINTER;
    }


    QString filename;

    // If its not just a buffer imagefile then set the default filename to be
    // its existing filename.
    if (!(*oldImageFile)->getFeatures().testFlag(ImageFile::F_Buffer))
        filename = getSaveAsFilename((*oldImageFile)->fileName());
    else
        filename = getSaveAsFilename("");

    qDebug() << "User chose to save as " << filename;

    long long expectedSize = (*oldImageFile)->device()->size();

    int retval = factory.createImage(filename,
                                     newImageFile,
                                     expectedSize,
                                     ImageFileFactory::F_DontCache);
    if (newImageFile == 0 || retval != errors::E_SUCCESS)  {
        qDebug() << "Failed to create new image file. Error code " << retval;
        return retval;
    }

    /*
     * Okay. We should have two image files now. oldImageFile and the
     * newImageFile the user just created. Now its time to copy all the
     * data from one to the other.
     */

    qDebug() << "Asking image files for their devices...";
    QIODevice *oldDevice = (*oldImageFile)->device();
    QIODevice *newDevice = (*newImageFile)->device();

    if (oldDevice == 0) {
        qDebug() << "ERROR: oldDevice is NULL.";
        return errors::E_BAD_POINTER;
    }
    if (newDevice == 0) {
        qDebug() << "ERROR: newDevice is NULL.";
        return errors::E_BAD_POINTER;
    }

    if (!(oldDevice->isOpen() && newDevice->isOpen())) {
        return errors::E_DEV_NOT_OPEN;
    }

    /*QProgressDialog qpd;
    qpd.setWindowTitle("Saving Disk Image...");
    qpd.setMaximum(expectedSize);

    long expectedBlocks = (long)(expectedSize / 512) + 1;
    long long bytesWritten = 0;

    for (int i = 0; i < expectedBlocks; i++) {
        qDebug() << "Copying block " << i;
        newDevice->write(oldDevice->read(512));
        bytesWritten += 512;
        qpd.setValue(bytesWritten);
    }
    qpd.setValue(expectedSize);*/

    qDebug() << "expectedSize: " << expectedSize;

    oldDevice->seek(0);
    newDevice->seek(0);

    newDevice->write(oldDevice->readAll());

    // See if we should set the comment.
    if ((*newImageFile)->getFeatures().testFlag(ImageFile::F_Comment)) {
        if ((*newImageFile)->getComment().isEmpty()) {
            QScopedPointer<QSettings> settings;
            QString settingsFile;
            settingsFile = QCoreApplication::applicationName() + ".ini";
            qDebug() << "Checking for configuration file " << settingsFile;

            if (QFile::exists(settingsFile)) {
                // Load settings from there
                settings.reset(new QSettings(settingsFile, QSettings::IniFormat, 0));
            } else {
                settings.reset(new QSettings());
            }
            qDebug() << "Settings Loaded.";

            if (settings->value("FileHandling/AskForComment",false).toBool()) {

                CommentPrompt cp;

                if ((*oldImageFile)->getFeatures().testFlag(ImageFile::F_Comment)) {
                    // Transfer over the comment
                    cp.setComment((*oldImageFile)->getComment());
                }

                if (cp.exec() == CommentPrompt::Accepted)
                    (*newImageFile)->setComment(cp.getComment());
            }
        }
    }

    (*newImageFile)->save();

    return errors::E_SUCCESS;
}








