/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: commentprompt.h
         Start Date: Sometime 2009
       Last Changed: $LastChangedDate: 2011-04-09 00:17:11 +1200 (Sat, 09 Apr 2011) $
           Revision: $Revision: 295 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2009,2010
            License: GNU General Public License
   File Description: Various utility functions
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/

#ifndef UTILS_H
#define UTILS_H

#include <QByteArray>
#include <QStringList>
#include "constants.h"
#include "version.h"

// On windows, we are able to get a list of removable media drives from the OS.
#ifdef Q_OS_WIN32
// This stuff is to support autodetecting removable media drives on Windows systems.
#include "windows.h"
#endif

/** A collection of utility functions.
 */
namespace utils {
    /** Swaps the byte ordering of the supplied QByteArray and returns it.

      \param b The array of bytes to swap the ordering of
      \returns The input data with its byte ordering swapped.
     */
    QByteArray swapEndian(QByteArray b);

    /** Swaps the byte ordering of the input QByteArray and then converts it
      to a 32bit integer.
      \param b The data to convert to an integer. The conversion starts at
                byte 0 and runs for four bytes. For the conversion to be
                successfull, there must be at least four bytes in the array.
      \returns The 32bit integer representation of the input data.
     */
    int byteArrayToIntS(QByteArray b);

    /** Converts the input byte array to a 32bit integer.

      At least four bytes are required to perform the conversion. The
      conversion starts at byte 0.

      It is a convenience function that just calls
      byteArrayToInt(QByteArray,int) passing it the input data and an offset
      of 0.

      \param b The byte array to convert to an integer.
      \returns The 32bit integer representation of the input data.
     */
    int byteArrayToInt(QByteArray b);

    /** Converts the input byte array to a 32bit integer starting at the
      specified offset.

      \param b The byte array to convert
      \param offset The place in the array to start converting. This offset
                    must be at least four bytes before the end of the array.
      \returns A 32bit integer representation of the four bytes starting at the
               specified offset.
     */
    int byteArrayToInt(QByteArray b, int offset);

    /** Returns true if the specified bit is set in the supplied byte of data.

      \param byte The byte to get the status of a bit from
      \param bit The bit number to be checked.
      \returns True if the specified bit is set in the supplied byte.
     */
    bool getBit(char byte, int bit);

    /** Returns a series of bytes from the input QByteArray.

      This method will return a QByteArray containing the exact number of bytes
      specified by the length parameter. If the input QByteArray or range is
      smaller than the specified length, the output will be padded with null
      bytes at the end.

      \param data The source data to copy into the output.
      \param start The location in the input data array to start copying from.
      \param end The location to the input data array to stop copying and start
                 padding with null bytes. If the difference between start and
                 end is the same as length, no null-byte padding will occur.
      \param length The number of bytes to copy. If there are not enough bytes
                    in the specified range, the output will be padded with
                    null-bytes.
      \returns All bytes between start and end, padded out to length with null
                bytes if required.
     */
    QByteArray getBytes(QByteArray data, int start, int end, int length);

    /** Copies the specified range of bytes from the input data array to the
        output.

        The end location should not be beyond the end of the input QByteArray.

        \param data The data array to copy from
        \param start The location in the input data array to start copying from
        \param end The location in the input data array to stop copying.
        \returns All data between the start and end locations in the input data
                 array.
     */
    QByteArray getBytes(QByteArray data, int start, int end);

    /** Gets a list of floppy drives on this system.

      On win32 systems, if detect_floppy_drives is true, this function will
      return a list of all removable media drives on the system.

      \returns A list of floppy drives or removable media drives.
     */
    QStringList getFloppyDrives();

    /** Gets the applications version string. Version numbers are specified in
      constants.h
      \returns The applications version string.
     */
    QString versionString();
}
#endif // UTILS_H
