/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "biosparamblock.h"
#include "util.h"

#include <QtDebug>

#define LESS_THAN_ONE_CHECK(var,varName) if (var < 1) {qWarning("%s value is invalid: should be greater than zero", varName); lastReturnCode = Errors::BPB_FATAL; return lastReturnCode;}

using namespace FilesystemFAT;
using namespace FilesystemFAT::Errors;

BiosParamBlock::BiosParamBlock(QObject *parent) :
    QObject(parent)
{
    close();
}

ReturnCode BiosParamBlock::load(QIODevice *dev, int offset) {
    ioDevice = dev;
    this->offset = offset;
    open = true;
    return reloadCache();
}

void BiosParamBlock::close() {
    ioDevice = 0;
    open = false;

    cOEMID = "";
    cBytesPerSector = 0;
    cSectorsPerCluster = 0;
    cReservedSectors = 0;
    cNumberOfFATs = 0;
    cMaxRootEntries = 0;
    cTotalSectors = 0;
    cMediaDescriptor = 0;
    cSectorsPerFAT = 0;
    cSectorsPerTrack = 0;
    cNumberOfHeads = 0;
    cHiddenSectors = 0;
    cExt12PhysicalDriveNumber = 0;
    cExt12Reserved = 0;
    cExt12ChkdskRequested = false;
    cExt12SurfaceScanRequested = false;
    cExt12ExtendedBootSignature = 0;
    cExt12VolumeID = 0;
    cExt12VolumeLabel = "";
    cExt12FSType = "";
}

ReturnCode BiosParamBlock::reloadCache() {
    qDebug() << "Loading BIOS Parameter Block into cache...";

    if (ioDevice.isNull()) {
        qWarning() << "BAD POINTER in BiosParamBlock::reloadCache(). "
                "Closing...";
        close();
        lastReturnCode = Errors::BAD_PTR;
        return lastErr();
    }

    qint64 cpos = ioDevice->pos();

    ioDevice->seek(offset);

    // 512 bytes should be enough.
    QByteArray data = ioDevice->read(512);

    if (data.length() < 512) {
        qWarning() << "Failed to read volume boot record. Aborting.";
        lastReturnCode = Errors::BPB_FATAL;
        return lastReturnCode;
    }

    ioDevice->seek(cpos);

    cOEMID = QString(Util::getBytes(data,3,10));

    // OEM_ID  -- ASCII String -- bytes 3 to 10
    cOEMID = QString(Util::getBytes(data,3,10));

    // BytesPerSector -- integer -- bytes 11 and 12
    cBytesPerSector = Util::byteArrayToIntS(Util::getBytes(data,11,12,4));

    // SectorsPerCluster -- integer -- byte 13
    cSectorsPerCluster = Util::byteArrayToIntS(Util::getBytes(data,13,13,4));

    // ReservedSectors -- integer -- bytes 14 and 15
    cReservedSectors = Util::byteArrayToIntS(Util::getBytes(data,14,15,4));

    // TotalFATs -- integer -- byte 16
    cNumberOfFATs = Util::byteArrayToIntS(Util::getBytes(data,16,16,4));

    // MaxRootEntries -- integer -- bytes 17 and 18
    cMaxRootEntries = Util::byteArrayToIntS(Util::getBytes(data,17,18,4));

    // TotalSectors -- integer -- bytes 19 and 20
    cTotalSectors = Util::byteArrayToIntS(Util::getBytes(data,19,20,4));

    // MediaDescriptor -- integer -- byte 21
    cMediaDescriptor = data.at(21);

    // SectorsPerFAT -- integer -- bytes 22 and 23
    cSectorsPerFAT = Util::byteArrayToIntS(Util::getBytes(data,22,23,4));

    // SectorsPerTrack -- integer -- bytes 24 and 25
    cSectorsPerTrack = Util::byteArrayToIntS(Util::getBytes(data,24,25,4));

    // NumHeads -- integer -- bytes 26 and 27
    cNumberOfHeads = Util::byteArrayToIntS(Util::getBytes(data,26,27,4));

    // HiddenSectors -- integer -- bytes 28 to 31
    cHiddenSectors = Util::byteArrayToIntS(Util::getBytes(data,28,31,4));

    // TotalSectorsLarge -- integer -- bytes 32 to 35
    int longTotalSectors = Util::byteArrayToIntS(Util::getBytes(data,32,35,4));

    // DriveNumber -- integer -- byte 36
    cExt12PhysicalDriveNumber = data.at(36);

    // Reserved (Windows NT Flags mostly) -- integer -- byte 37
    cExt12Reserved = data.at(37);

    // Signature -- integer -- byte 38
    cExt12ExtendedBootSignature = data.at(38);

    // VolumeID -- integer -- bytes 39 to 42
    cExt12VolumeID = Util::byteArrayToIntS(Util::getBytes(data,39,42,4));

    // VolumeLabel -- string -- bytes 43 to 53
    cExt12VolumeLabel = QString(Util::getBytes(data,43,53));

    // SystemID -- string == bytes 54 to 61
    cExt12FSType = QString(Util::getBytes(data,54,61));

    // If thetwo byte value at offset 0x13 is empty then the total number of
    // sectors is greater than 65535 so we should use the four byte value at
    // 0x20.
    if (cTotalSectors == 0) {
        qDebug() << "Value at offset 0x13 is empty. Using value at offset 0x20 "
                "for Total Sectors instead.";
        cTotalSectors = longTotalSectors;
    }

    //-----------
    // Do some quick checks on any values we can.

    if (cExt12ExtendedBootSignature != 0x29) {
        qDebug() << "There is probably no FAT12/16 Extended Boot Signature.";
    }

    // Check if sectorsPerCluster is a power of two
    if (!((cSectorsPerCluster != 0) &&
          ((cSectorsPerCluster & (cSectorsPerCluster - 1)) == 0))) {
        qWarning() << "SectorsPerCluster value is invalid: Should be a power "
                "of two.";
        lastReturnCode = Errors::BPB_WARNING;
    }

    // Ensure some important values are one or greater (to prevent division
    // by one errors and such)
    LESS_THAN_ONE_CHECK(cMaxRootEntries,"MaxRootEntries")
    LESS_THAN_ONE_CHECK(cBytesPerSector,"BytesPerSector")
    LESS_THAN_ONE_CHECK(cNumberOfFATs,"Number of File Allocation Tables")
    LESS_THAN_ONE_CHECK(cSectorsPerFAT, "SectorsPerFAT")
    LESS_THAN_ONE_CHECK(cSectorsPerCluster, "SectorsPerCluster")

    // Check that sectorsPerCluster value is not such that bytes per cluster
    // is greater than 32KB
    if ((cBytesPerSector * cSectorsPerCluster) > 32768) {
        qWarning() << "SectorsPerCluster or BytesPerSector value is probably "
                "invalid: Bytes Per Cluster should not exceed 32KB";
        lastReturnCode = Errors::BPB_WARNING;
    }

    // Check that root directory size is a multiple of the sector size
    if (((32*cMaxRootEntries) % cBytesPerSector) != 0) {
        qWarning() << "Root Directory size is not a multiple of the "
                "sector size";
        lastReturnCode = Errors::BPB_WARNING;
    }

    // Check that we have at least one file allocation table. This is the
    // filesystems core data structure and we can't do anything without it.
    if (numberOfFATs() < 1) {
        qCritical() << "Invalid FAT count. There must be at least ONE File "
                "Allocation Table for the filesystem to be loaded.";
        lastReturnCode = Errors::FAT_NO_FAT;
    }

    // If there weren't any (possibly recoverable) errors, then we must have
    // been successfull. If there were unrecoverable errors we wouldn't have
    // got this far.
    if (lastReturnCode != Errors::BPB_WARNING &&
        lastReturnCode != Errors::FAT_NO_FAT)
        lastReturnCode = Errors::SUCCESS;
    return lastErr();
}

QString BiosParamBlock::oemID() const {
    return cOEMID;
}

unsigned int BiosParamBlock::bytesPerSector() const {
    return cBytesPerSector;
}

unsigned int BiosParamBlock::sectorsPerCluster() const {
    return cSectorsPerCluster;
}

unsigned int BiosParamBlock::reservedSectors() const {
    return cReservedSectors;
}

unsigned int BiosParamBlock::numberOfFATs() const {
    return cNumberOfFATs;
}

unsigned int BiosParamBlock::maxRootEntries() const {
    return cMaxRootEntries;
}

unsigned long BiosParamBlock::totalSectors() const {
    return cTotalSectors;
}

unsigned char BiosParamBlock::mediaDescriptor() const {
    return cMediaDescriptor;
}

unsigned int BiosParamBlock::sectorsPerFAT() const {
    return cSectorsPerFAT;
}

unsigned int BiosParamBlock::sectorsPerTrack() const {
    return cSectorsPerTrack;
}

unsigned int BiosParamBlock::numberOfHeads() const {
    return cNumberOfHeads;
}

unsigned long BiosParamBlock::hiddenSectors() const {
    return cHiddenSectors;
}

unsigned char BiosParamBlock::extPhysicalDriveNumber() const {
    return cExt12PhysicalDriveNumber;
}

unsigned char BiosParamBlock::ext12Reserved() const {
    return cExt12Reserved;
}

bool BiosParamBlock::ext12ChkdskRequested() const {
    return cExt12ChkdskRequested;
}

bool BiosParamBlock::ext12SurfaceScanRequested() const {
    return cExt12SurfaceScanRequested;
}

unsigned char BiosParamBlock::ext12ExtendedBootSignature() const {
    return cExt12ExtendedBootSignature;
}

unsigned int BiosParamBlock::ext12VolumeID() const {
    return cExt12VolumeID;
}

QString BiosParamBlock::ext12VolumeLabel() const {
    return cExt12VolumeLabel;
}

QString BiosParamBlock::ext12FSType() const {
    return cExt12FSType;
}
