/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "datachain.h"
#include "fatfs.h"
#include <QtDebug>

using namespace FilesystemFAT;

DataChain::DataChain(FatFS *fs,
                     unsigned int firstCluster,
                     QObject *parent) :
    QObject(parent)
{
    qDebug() << "----> DataChain::DataChain(AbstractFAT*,DataArea*,"
                "BiosParamBlock*,uint,QObject*)";

    /* Store useful pointers */
    this->fs = fs;
    this->data = fs->getDataArea();
    this->bpb = fs->getBPB();

    // Load up the cluster chain.
    ReturnCode rc = init(firstCluster);
    if (rc != Errors::SUCCESS) {
        lastErr = rc;
        return;
    }

    qDebug() << "Calculating length...";

    // Figure out the length based on the number of clusters.
    mLength = clusterChain.length() * (bpb->bytesPerSector() *
                                       bpb->sectorsPerCluster());
    seek(0);
    lastErr = Errors::SUCCESS;
    qDebug() << "DataChain init completed.";
}

DataChain::DataChain(FatFS *fs,
                     unsigned int firstCluster,
                     unsigned int length,
                     QObject *parent):
    QObject(parent)
{
    qDebug() << "----> DataChain::DataChain(AbstractFAT*,DataArea*,"
                "BiosParamBlock*,uint,uint,QObject*)";

    /* Store useful pointers */
    this->fs = fs;
    this->data = fs->getDataArea();
    this->bpb = fs->getBPB();

    // Load up the cluster chain.
    ReturnCode rc = init(firstCluster);
    if (rc != Errors::SUCCESS) {
        lastErr = rc;
        return;
    }

    // We can't really rely on the directory entry reporting the length
    // correctly. 99% of the time it should be fine but there is always the
    // possibility of corruption.
    unsigned int computedLength =clusterChain.length() *
                                 (bpb->bytesPerSector() *
                                  bpb->sectorsPerCluster());

    // We can't possibly read more than the computed length. If the reported
    // length is larger then it is obviously wrong so we'll use the computed
    // one.
    if (computedLength < length) {
        qWarning() << "Length reported by Directory Entry seems to be "
                "incorrect. Using computed length instead. This may result in "
                "some garbage at the end of the file.";
        this->mLength = computedLength;
    }
    else
        this->mLength = length;
    seek(0);
    lastErr = Errors::SUCCESS;
    qDebug() << "DataChain init completed.";
}

Errors::ReturnCode DataChain::init(unsigned int firstCluster) {

    // Try to read the data chain from the first FAT.
    AbstractFAT* fat = fs->getFAT(0);
    clusterChain = fat->getChain(firstCluster);

    if (fat->lastError() != Errors::SUCCESS) {
        // Failed. Time to try any other FATs.
        unsigned int fatCount = bpb->numberOfFATs();
        if (fatCount > 1) {
            for (unsigned int i = 1; i < fatCount; i++) {
                qWarning() << "Error detected reading from FAT #" << i-1
                           << ", retrying with FAT #" << i;
                fat = fs->getFAT(i);
                clusterChain = fat->getChain(firstCluster);

                if (fat->lastError() == Errors::SUCCESS) {
                    qDebug() << "Succeeded reading from FAT after" << i+1
                             << "attempts";
                    return Errors::SUCCESS;
                }
            }
            qWarning() << "Failed to load cluster chain after" << fatCount
                       << "attempts. Aborting.";
            return fat->lastError();
        } else {
            qWarning() << "Error reading cluster chain. Aborting.";
            lastErr = fat->lastError();
            seek(0);
            return Errors::SUCCESS;
        }
    }
    return Errors::SUCCESS;
}

bool DataChain::seek(unsigned int pos) {
    qDebug() << "----> DataChain::seek(" << pos << ")";

    /* Can't seek past the end of the file
     */
    if (pos > length()) {
        lastErr = Errors::SEEK_PAST_END;
        return false;
    }

    /* See if we are at the end of the file */
    if (pos == length()) isAtEnd = true;
    else isAtEnd = false;

    pointer = pos;
    lastErr = Errors::SUCCESS;
    return true;
}

unsigned int DataChain::translateClusterNumber(int localCluster) {
    qDebug() << "Translate Cluster Number" << localCluster;
    if (localCluster > clusterChain.length()) {
        lastErr = Errors::INVALID_CLUSTER;
        return 0;
    }

    return clusterChain.at(localCluster);
}

QByteArray DataChain::read(unsigned int length) {
    qDebug() << "----> DataChain::read(" << length << ")";
    qDebug() << "------------------------------------------------------------";
    qDebug() << "Reading" << length << "bytes from position" << pointer;
    QByteArray buffer;

    /* First up: compute the _real_ length */
    if (pointer + length > this->length()) {
        /* Cant read past end of file. We will read to the end instead. */
        qWarning() << "Can't read past end of file. Reading to end instead.";
        length = this->length() - pointer;
    }

    /* Need to know how many bytes per cluster. */
    const unsigned int bytesPerCluster = bpb->bytesPerSector()
                                         * bpb->sectorsPerCluster();
    qDebug() << "Bytes per cluster:" << bytesPerCluster;

    /* Figure out which cluster we are starting in */
    const unsigned int firstCluster = pointer / bytesPerCluster;
    qDebug() << "First Cluster:" << firstCluster;

    /* And which cluster we are ending in */
    const unsigned int lastCluster = (pointer + length) / bytesPerCluster;
    qDebug() << "Last Cluster:" << lastCluster;

    /* How far into the first cluster we should start reading */
    const unsigned int firstByte = pointer - (firstCluster * bytesPerCluster);
    qDebug() << "Start position in first cluster:" << firstByte;

    /* And how far into the last cluster we should stop reading */
    const unsigned int lastByte = (pointer + length) - (lastCluster
                                                        * bytesPerCluster);
    qDebug() << "End position in last cluster:" << lastByte;

    /* Read in the first cluster. We will need it regardless of what happens. */
    QByteArray tempData = data->readCluster(
                                        translateClusterNumber(firstCluster));

    /* See if we have to look at more than one cluster */
    if (firstCluster < lastCluster) {
        /* Append data from the first cluster */
        buffer.append(tempData.mid(firstByte));
        qDebug() << "Appended end of cluster" << firstCluster;

        /* Now we get the data for all clusters _between_ the first and last.
           We skip over this if the last cluster directly follows the first. */
        if (lastCluster > firstCluster + 1) {
            for (unsigned int i = firstCluster + 1; i < lastCluster; i++) {
                buffer.append(data->readCluster(translateClusterNumber(i)));
            }
        }

        /* Now deal with the last cluster. If the last byte is zero then there
           is no point in looking at the last cluster as we don't even want
           anything from it.
         */
        if (lastByte > 0) {
            tempData.clear();
            qDebug() << "Appending" << lastByte << "bytes at start of final cluster";
            tempData = data->readCluster(translateClusterNumber(lastCluster));
            buffer.append(tempData.mid(0,lastByte));
            qDebug() << "Appended start of cluster" << lastCluster;
        }
    } else {
        /* The first and end cluster are the same. We are only reading a bit
           from inside a single cluster. */
        const unsigned int readLength = firstByte - lastByte;
        qDebug() << "Appended middle of cluster" << firstCluster;
        buffer.append(tempData.mid(firstByte,readLength));
    }

    qDebug() << "------------------------------------------------------------";

    lastErr = Errors::SUCCESS;
    return buffer;
}

QByteArray DataChain::readAll() {
    qDebug() << "----> DataChain::readAll()";
    seek(0);
    return read(length());
}
