/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef DATACHAIN_H
#define DATACHAIN_H

#include <QObject>
#include <QPointer>
#include <QByteArray>

#include "abstractfat.h"
#include "dataarea.h"
#include "biosparamblock.h"
#include "errors.h"

namespace FilesystemFAT {

    class FatFS;

    /** DataChain represents File/Directory data in the FAT filesystem.

        In the FAT filesystem, file (and directory) data is stored as a linked
        list. The links are stored in the File Allocation Table and the Data is
        stored in the Data Area. Each entry in the FAT matches a cluster. So to
        look up the next cluster for a file you look at the entry in the FAT
        for the current cluster. This entry gives you the next cluster.

        DataChain abstracts all of that away. It takes a class implementing
        AbstractFAT and the DataArea and provides random access to file data.
     */
    class DataChain : public QObject
    {
    Q_OBJECT
    public:
        /** Creates a new data chain. As no length is specified the length is
            calculated as the number of clusters it occupies.
          \param fs The Filesystem Instance
          \param firstCluster The first cluster in the data chain.
          \param parent The parent QObject.
         */
        explicit DataChain(FatFS *fs,
                           unsigned int firstCluster,
                           QObject *parent = 0);

        /** Creates a new data chain. This constructor allows a maximum length
            to be specified. If this is larger than the calculated maximum size
            (number of clusters * cluster size) then then the calculated
            maximum is used instead.
          \param fs The Filesystem instance
          \param firstCluster The first cluster in the data chain.
          \param length The length of the DataChain.
          \param parent The parent QObject.
         */
        explicit DataChain(FatFS *fs,
                           unsigned int firstCluster,
                           unsigned int length,
                           QObject *parent = 0);

        /** Returns the last error code to occur.
          \returns The last error code to occur.
         */
        ReturnCode lastError() {return lastErr;}

        /* ***************************************************************** *
         * Note that these doxygen comments are copied into FsFile           *
         * automatically. Avoid any references to DataChain in them - they   *
         * wont make sense in the FsFile documentation.                      *
         * ***************************************************************** */

        /** Changes the location pointer to the one supplied.
          \param pos The new position for the location pointer.
          \returns True if the seek was successfull.
         */
        bool seek(unsigned int pos);

        /** Returns the current location pointer value.
          \returns The current location pointer value.
         */
        unsigned int pos() {return pointer;}

        /** Returns the length of the file.
          \returns The length of the file.
         */
        unsigned int length() {return mLength;}

        /** Reads at most length bytes and returns it in a QByteArray.
          The location pointer is incremented after the read is completed.
          \param length The number of bytes to read.
          \returns A QByteArray containing at most length bytes from the
          current location pointer.
         */
        QByteArray read(unsigned int length);

        /** Reads the entire file and returns it in a QByteArray.
          \returns The entire file.
         */
        QByteArray readAll();

        /** Returns true if currently at the end of the file.
          \returns True if currently at the end of the file.
         */
        bool atEnd() {return isAtEnd;}

    private:
        /** Translates a local cluster number (a cluster for this file) into a
            global cluster number (the cluster number for the entire disk).
            This essentially just returns clusterChain.at(localCluster).
          \param localCluster The file cluster number we want the data area
                 cluster number for
          \returns The Data Area cluster number.
         */
        unsigned int translateClusterNumber(int localCluster);

        /** Load up the cluster chain, etc.
          \param firstCluster The first cluster in the chain.
         */
        Errors::ReturnCode init(unsigned int firstCluster);

        QPointer<FatFS> fs;                 /*!< The Filesystem instance */
        QPointer<DataArea> data;            /*!< The Data Area we are using */
        QPointer<BiosParamBlock> bpb;       /*!< The BIOS Parameter Block to
                                                 get info from */
        QList<unsigned int> clusterChain;   /*!< First cluster */
        unsigned int mLength;               /*!< Length of the file */
        unsigned int pointer;               /*!< Current file pointer */
        bool isAtEnd;                       /*!< If we are at the end of the
                                                 file or not */
        ReturnCode lastErr;                 /*!< Last error code */
    };
}

#endif // DATACHAIN_H
