/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "directory.h"
#include "constants.h"
#include "fatfs.h"
#include "lfncomponent.h"

#include <QStringList>
#include <QtDebug>

using namespace FilesystemFAT;

// Constructs the root directory
Directory::Directory(FatFS * filesystem,
                     QObject *parent) :
    DirectoryEntry(parent)
{
    qDebug() << "----> Directory::Directory(FatFS*,QObject*)";

    qDebug() << "Creating a root directory.";

    this->filesystem = filesystem;

    this->parentPathName = "/";
    this->fullPathName = "/";

    // load directory entries from disk.
    refresh();
}

// Constructs a subdirectory
Directory::Directory(FatFS *filesystem,
                     QByteArray entryData,
                     QString parentPathName,
                     QObject *parent) :
    DirectoryEntry(entryData,parent) {
    qDebug() << "----> Directory::Directory(FatFS*,QByteArray,parent:"
             << parentPathName << ",QObject*)";

    qDebug() << "Creating a Subdirectory.";

    this->filesystem = filesystem;

    // Ensure it ends with a slash
    if (!parentPathName.endsWith("/")) parentPathName += "/";

    this->parentPathName = parentPathName;
    this->fullPathName = parentPathName + fileName() + "/";

    // Load directory entries from disk.
    refresh();
}

// Constructs a subdirectory
Directory::Directory(FatFS *filesystem,
                     QByteArray entryData,
                     QString parentPathName,
                     QString longFilename,
                     QObject *parent) :
    DirectoryEntry(entryData,parent) {
    qDebug() << "----> Directory::Directory(FatFS*,QByteArray,parent:"
             << parentPathName << ",QObject*)";

    qDebug() << "Creating a Subdirectory.";

    this->filesystem = filesystem;

    // Ensure it ends with a slash
    if (!parentPathName.endsWith("/")) parentPathName += "/";

    this->parentPathName = parentPathName;
    this->fullPathName = parentPathName + fileName() + "/";
    this->mLongFilename = longFilename;

    // Load directory entries from disk.
    refresh();
}

Directory::~Directory() {
    qDebug() << "Destroying instance of directory" << fullPathName;

    clearDirEntries();
}

QList<DirectoryPtr> Directory::getSubdirs() {
    qDebug() << "----> Directory::getSubdirs()";
    qDebug() << "Getting subdirectory list...";
    qDebug() << "There are" << dirents.count() << "Directory Entries.";
    QList<DirectoryPtr> dirList;

    for (int i = 0; i < dirents.count(); i++) {

        DirectoryEntry* current = dirents.at(i);

        if (current->isSubdirectory()) {
            /* Creates a new directory with the FatFS instance as its parent.
               It passes through the FatFS instance, entry data and this
               directories name
             */
            DirectoryPtr dir(new Directory(filesystem.data(),
                                           current->getEntryData(),
                                           fullPathName,
                                           current->mLongFilename,
                                           filesystem.data()));
            dirList.append(dir);
        }
    }

    lastErr = Errors::SUCCESS;
    return dirList;
}

Directory * Directory::getSubdir(QString directoryName) {
    qDebug() << "----> Directory::getSubdir(" << directoryName << ")";

    qDebug() << "In" << fullPathName
             << "Searching for subdirectory" << directoryName;
    qDebug() << "There are" << dirents.count() << "Directory Entries.";

    for (int i = 0; i < dirents.count(); i++) {
        DirectoryEntry * current = dirents.at(i);

        // FAT is case insensitive so we convert to uppercase when comparing
        // entries.
        if (current->isSubdirectory() &&
            current->fileName().toUpper() == directoryName.toUpper()) {

            qDebug() << "Found matching directory entry.";

            /* Creates a new directory with the FatFS instance as its parent.
               It passes through the FatFS instance, entry data and this
               directories name
             */
            lastErr = Errors::SUCCESS;
            return new Directory(filesystem.data(),
                                 current->getEntryData(),
                                 fullPathName,
                                 current->mLongFilename,
                                 filesystem.data());
        }
    }
    qDebug() << "Subdirectory not found.";
    lastErr = Errors::FILE_NOT_FOUND;
    return NULL;
}

FsFile * Directory::getFile(QString fileName) {
    qDebug() << "----> Directory::getFile(" << fileName << ")";
    qDebug() << "In" << fullPathName << "Searching for file" << fileName;
    qDebug() << "There are" << dirents.count() << "Directory Entries.";

    for (int i = 0; i < dirents.count(); i++) {
        DirectoryEntry * current = dirents.at(i);

        // FAT is case insensitive so we convert to uppercase when comparing
        // entries.
        if (!current->isSubdirectory() &&
            current->fileName().toUpper() == fileName.toUpper()) {

            qDebug() << "Found matching directory entry.";

            /* Create a new FsFile object passing in its entry data,
               a pointer to the FatFS instance (so it can get at the FAT, etc),
               the full pathname of this object (so it can get its parent, etc),
               and a pointer to the FatFS instance again - this time as the
               parent QObject.
             */
            lastErr = Errors::SUCCESS;
            return new FsFile(current->getEntryData(),
                              filesystem.data(),
                              fullPathName,
                              current->mLongFilename,
                              filesystem.data());
        }
    }
    qDebug() << "file not found.";
    lastErr = Errors::FILE_NOT_FOUND;
    return NULL;
}

QList<FsFilePtr> Directory::getFiles() {
    qDebug() << "----> Directory::getFiles()";

    qDebug() << "Getting file list...";
    QList<FsFilePtr> fileList;

    for (int i = 0; i < dirents.count(); i++) {
        DirectoryEntry * current = dirents.at(i);
        if (!current->isSubdirectory()) {

            /* Create a new FsFile object passing in its entry data,
               a pointer to the FatFS instance (so it can get at the FAT, etc),
               the full pathname of this object (so it can get its parent, etc),
               and a pointer to the FatFS instance again - this time as the
               parent QObject.
             */
            FsFilePtr file(new FsFile(current->getEntryData(),
                                      filesystem.data(),
                                      fullPathName,
                                      current->mLongFilename,
                                      filesystem.data()));
            fileList.append(file);
        }
    }

    return fileList;
}

Directory *Directory::getParent() {
    qDebug() << "----> Directory::getParent()";
    if (!filesystem.isNull())
        return filesystem->createDirectory(parentPathName);
    else {
        qCritical() << "Filesystem object is NULL. Unable to get parent "
                "directory. Returning null.";
        return NULL;
    }
}

Directory *Directory::getDirectory(QString directoryName) {
    qDebug() << "----> Directory::getDirectory(" << directoryName << ")";

    qDebug() << "Directory" << directoryName << "requested. Searching... ";

    /* If the directoryName starts with either our name, a ./ sequence or
     * no / character then it is under our level somewhere. We will try to
     * process it locally rather than handing it up to the FatFS instance.
     * That should reduce the number of Directories that have to be visited
     * to find the right one.
     */

    if (directoryName == ".") {
        /* We are being asked for this directory. Its safer to get the FatFS to
         * do that rather than attempt to clone ourself.
         */

        qDebug() << "Handing off to Filesystem instance (self requested)";

        return getDirFromFS(fullPathName);
    }

    if (directoryName == "..") {
        /* The parent directory was asked for. That is one above our current
         * level so we will have to ask the filesystem instance to create it
         * for us.
         */

        qDebug() << "Parent directory requested. Handing off to Filesystem "
                "instance.";

        return getDirFromFS(parentPathName);
    }

    if (directoryName.startsWith("/")) {
        if (directoryName.toUpper().startsWith(fullPathName.toUpper())) {
            /* The directory name starts with the full name of this directory.
             * So it is somewhere under our level. We should start here.
             */

            qDebug() << "Reqested directory starts with current pathname, "
                    "handling locally.";

            return findSubdirectory(directoryName);
        } else {
            /* The directory name starts with a "/" (so the path starts in the
             * root directory) and does not start with this directories full
             * path name. So it is not somewhere under this directory - hand it
             * off to FatFS.
             */

            qDebug() << "Requested directory is outside of this one. Handing "
                    "off to Filesystem instance.";

            return getDirFromFS(directoryName);
        }
    }

    if (directoryName.startsWith("./") ||
        !directoryName.startsWith("/")) {
        /* It starts in the current directory. Chop it off and do a local search
         * for the directory.
         */

        qDebug() << "Handling locally.";

        if (directoryName.startsWith("./"))
            return findSubdirectory(directoryName.mid(2));
        else
            return findSubdirectory(directoryName);
    }


    qDebug() << "Unable to decide what to do. Handing off to Filesystem "
            "instance.";
    return getDirFromFS(directoryName);
}

void Directory::refresh() {
    qDebug() << "----> Directory::refresh()";

    qDebug() << "Loading directory data...";

    QByteArray directoryData;
    unsigned int maxEntries;

    if (filesystem.isNull()) {
        qCritical() << "Filesystem object is NULL. Unable to proceed with "
                "directory refresh.";
        return;
    }

    QPointer<QIODevice> device = filesystem->getDevice();
    QPointer<BiosParamBlock> bpb = filesystem->getBPB();
    QPointer<AbstractFAT> fat = filesystem->getFAT();
    QPointer<DataArea> dataArea = filesystem->getDataArea();

    bool lfnEnabled = filesystem->isLFNSupportEnabled();

    if (device.isNull() || bpb.isNull() || fat.isNull() || dataArea.isNull()) {
        qCritical() << "Null Pointer in Directory Refresh. Aborting.";
        lastErr = Errors::BAD_PTR;
        return;
    }

    // First up read the directory data.
    if (fullPathName == "/") {
        qDebug() << "Loading as a Root Directory.";
/*  We have a QIODevice set. That means we must be supposed to load the root
    directory. That lives out on its own. Here is how we calculate its location:
        _______________________FAT 1_______________________  ________________________FAT 2________________________
       /                                                   \/                                                     \
            1     2     3     4     5     6     7     8     9     1     2     3     4     5     6     7     8     9
+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+---...
| RES |         File Allocation Table Number 1 of 2         |        File Allocation Table Number 2 of 2          | Root Directory...
+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+---...
0     1     2     3     4     5     6     7     8     9     10    11    12    13    14    15    16    17    18    19    20
      Sector Number

    The number of sectors per FAT and number of bytes per sector are stored in
    the BIOS Parameter Block (that first REServed sector).
*/
        const unsigned int rootDirStart =
                (bpb->reservedSectors() + /* Boot sector(s) */
                (bpb->numberOfFATs() * bpb->sectorsPerFAT()) /* FATs */
                )* bpb->bytesPerSector(); /* convert to bytes */
        const unsigned int rootDirLength = bpb->maxRootEntries() *
                                           Constants::bytesPerDirEntry;

        qDebug() << "Loading root directory from byte:" << rootDirStart;
        qDebug() << "Root directory length:" << rootDirLength;

        qint64 pos = device->pos();
        device->seek(rootDirStart);
        directoryData = device->read(rootDirLength);
        device->seek(pos);

        maxEntries = bpb->maxRootEntries();
    } else {
        qDebug() << "Loading as a Subdirectory.";

        // It is a subdirectory. Use a data chain to load the data.
        DataChain dc(filesystem.data(),entryCluster(),this);

        // If something went wrong with the data chain, exit.
        if (dc.lastError() != Errors::SUCCESS) {
            lastErr = dc.lastError();
            clearDirEntries();
            return;
        }

        directoryData = dc.readAll();

        maxEntries = directoryData.size() / Constants::bytesPerDirEntry;
    }

    // Wipe out the existing list of directory entries as we are about to
    // reload them
    clearDirEntries();


    QList<LfnComponent *> lfnComponents;

    // Parse out the directory data.
    for (unsigned int i = 0; i < maxEntries; i++) {
        // Grab the directory entry
        QByteArray direntData = directoryData.mid(i *
                                                  Constants::bytesPerDirEntry,
                                                  Constants::bytesPerDirEntry);

        // Is this really a directory entry? Sure it isn't part of some files
        // long filename?
        if (direntData.at(0x0B) == 0x0F) {
            if (lfnEnabled) {
                // Turns out it is. We will create this as a LfnComponent
                // instead.
                LfnComponent *c = new LfnComponent(direntData,this);

                // And put it aside for later.
                lfnComponents.append(c);
            }
            continue;
        }

        DirectoryEntry * ent = new DirectoryEntry(direntData,this);

        QString fn = ent->fileName();

        // Check to see if this is a valid directory entry
        if (ent->isVolumeLabel()) {
            volumeLabel = fn;
        } else if (fn.trimmed() == "."  || /* Pointer to current directory */
                   fn.trimmed() == ".." || /* Pointer to parent directory */
                   fn.isEmpty()         || /* No filename, no file */
                   ent->isDeleted()        /* File has been deleted */
                   ) {
            //  Ignore the entry.
        } else if (ent->isFinalEntry()) {
            // This is the end of the directory.
            break;
        }else {
            qDebug() << "Entry accepted: " << fn;
            dirents.append(ent);
        }
    }

    if (lfnEnabled) {
        qDebug() << "========================================================";
        qDebug() << "Processing Long Filename chunks...";
        qDebug() << "There are" << lfnComponents.count()
                 << "Long filename components.";
        bool deFound = false;
        // If we have any Long Filename components to deal with, do so now.
        if (!lfnComponents.isEmpty()) {
            while (!lfnComponents.isEmpty()) {
                LfnComponent *curLFN = lfnComponents.takeFirst();

                for (int j = 0; j < dirents.count(); j++) {
                    DirectoryEntry *de = dirents.at(j);

                    if (curLFN->getDOSFilenameChecksum() ==
                                                   de->getFilenameChecksum()) {

                        // Pass the component off to the directory entry. It
                        // will take ownership of it so we never need to deal
                        // with it again.
                        de->addLFNComponent(curLFN);

                        // Make a note that we found someone to take the long
                        // filename component
                        deFound = true;
                        break;
                    }
                }

                // If we found a directory entry to take it.
                if (!deFound) {
                    qDebug() << "Discarding LFN component:"
                             << curLFN->getFilenameData();
                    delete curLFN;
                }
            }
        }
        qDebug() << "========================================================";
    }

    lastErr = Errors::SUCCESS;
}

Directory *Directory::findSubdirectory(QString directoryName) {
    qDebug() << "----> Directory::findSubdirectory(" << directoryName << ")";

    qDebug() << "Searching locally for directory " << directoryName;

    // This is the path we must follow
    QStringList path = directoryName.split('/');

    // Clear out blank entries.
    for (int i = 0; i < path.count(); i++)
        if (path.at(i).isEmpty())
            path.removeAt(i);

    if (path.length() < 1) {
        lastErr = Errors::FILE_NOT_FOUND;
        return NULL;
    }

    // We'll start right here.
    Directory * prevdir = NULL;
    Directory * curdir = this;

    // start searching
    while (!path.isEmpty()) {
        // Grab the next directory name
        QString nextDirectory = path.takeFirst();

        // An empty directory name is probably just a typo (putting // instead
        // of /) - continue.
        if (nextDirectory.isEmpty()) continue;

        prevdir = curdir;
        curdir = prevdir->getSubdir(nextDirectory);
        if (curdir == NULL) {
            /* getSubdir returned null - the subdirectory couldnt be
               found. That means that the directory we are searching for must
               not exist. Return null to indicate this.
             */
            lastErr = Errors::FILE_NOT_FOUND;
            return NULL;
        }

        if (!path.isEmpty() && prevdir != this) {
            /* If we aren't currently looking at ourself OR the last directory
             * then we don't care about it. Throw it away.
             *
             * (We obviously can't delete ourself or the directory we are
             * actually looking for - that would be silly)
             */

            delete prevdir;
        }
    }

    if (curdir == this) {
        // Wee need another copy of ourself
        return getDirFromFS(fullPathName);
    }


    lastErr = Errors::SUCCESS;
    return curdir;
}

Directory *Directory::getDirFromFS(QString directoryName) {
    qDebug() << "----> Directory::getDirFromFS(" << directoryName << ")";
    qDebug() << "Requesting directory from FatFS instance: " << directoryName;

    Directory *tmp = filesystem->createDirectory(directoryName);
    if (filesystem->lastError() == Errors::FILE_NOT_FOUND) {
        lastErr = Errors::FILE_NOT_FOUND;
        return NULL;
    } else {
        lastErr = Errors::SUCCESS;
        return tmp;
    }

    return NULL;
}

void Directory::clearDirEntries() {
    qDebug() << "----> Directory::clearDirEntries()";

    qDebug() << "Clearing cached directory entries.";
    for (int i = 0; i < dirents.count(); i++) {
        delete dirents.takeFirst();
    }
}
