/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef DIRECTORY_H
#define DIRECTORY_H

#include <QSharedPointer>
#include <QPointer>
#include <QList>

#include "directoryentry.h"
#include "datachain.h"
#include "abstractfat.h"
#include "biosparamblock.h"
#include "fsfile.h"

namespace FilesystemFAT {

    class Directory;
    class FatFS;

    /** A shared pointer for a Directory.
     */
    typedef QSharedPointer<Directory> DirectoryPtr;

    /** The directory object handles directories in the FAT Filesystem. There
        are two types of directories:
         - The FAT12/16 Root directory. This is stored out side of the data
           area and is handled specially. It has no parent directory and no
           attributes aside from the filename (which actually returns the
           volume label as the root). It is constructed using the constructor
           which does not take entryData.
         - Subdirectories. These are constructed using a 32-byte directory
           entry.

        Anything that returns a list of directory entries returns them wrapped
        in a QSharedPointer. Methods which return a single entry return them as
        pointers.
     */
    class Directory : public DirectoryEntry
    {
    Q_OBJECT
    public:
        /** Constructs the Root Directory. It passes the string "/" (plus the
            entries name) to all subdirectories and files it creates for the
            fullPathName.

            FAT12/16 stores the root directory outside of the data area
            separately from everything else. This requires it to be handled
            specially so it has its own constructor.

          \param filesystem The filesystem this directory is a part of.
          \param parent The parent QObject. This should be the FatFS instance.
         */
        explicit Directory(FatFS * filesystem,
                           QObject *parent = 0);

        /** Constructs a subdirectory. This can be any directory except the
            root directory as that must be handled specially.

          \param filesystem The FAT Filesystem this directory is a part of.
          \param entryData The directory entry data from the parent directory.
          \param parentPathName The parent directories full path name
                                (eg "/foo/bar/123/"). This is used to look up
                                the parent directory. It is passed on to any
                                file or directory objects created by this
                                object with this objects name appended. It
                                should have a trailing "/" character.
          \param parent The parent QObject. This should usually be a FatFS
                        instance.
         */
        explicit Directory(FatFS * filesystem,
                           QByteArray entryData,
                           QString parentPathName,
                           QObject *parent = 0);

        /** Constructs a subdirectory. This can be any directory except the
            root directory as that must be handled specially.

            This constructor takes the long filename (which isn't stored in
            the directory entry data).

          \param filesystem The FAT Filesystem this directory is a part of.
          \param entryData The directory entry data from the parent directory.
          \param parentPathName The parent directories full path name
                                (eg "/foo/bar/123/"). This is used to look up
                                the parent directory. It is passed on to any
                                file or directory objects created by this
                                object with this objects name appended. It
                                should have a trailing "/" character.
          \param longFilename The directories long filename.
          \param parent The parent QObject. This should usually be a FatFS
                        instance.
         */
        explicit Directory(FatFS *filesystem,
                           QByteArray entryData,
                           QString parentPathName,
                           QString longFilename,
                           QObject *parent);

        /** Cleans up the directory.
         */
        ~Directory();

        /** Gets a list of all subdirectories directly under this directory. To
          get a specific directory by name use getSubdir() instead.
          \returns A list of all subdirectories directly under this directory.
         */
        QList<DirectoryPtr> getSubdirs();

        /** Gets the specified directory. If the specified directory could not
         *  be found in this directory, null will be returned and an error
         *  condition set.
         * \param directoryName the name of the directory to get.
         * \returns The specified directory or null if it could not be found.
         */
        Directory * getSubdir(QString directoryName);

        /** Gets a list of all files directly under this directory.
          \returns A list of all files directly under this directory.
         */
        QList<FsFilePtr> getFiles();

        /** Gets the specified file. If the specified file could not be found
            in this directory, null will be returned and an error condition set.
          \param fileName the name of the file to get.
          \returns The specified file or null if it could not be found.
         */
        FsFile * getFile(QString fileName);

        /** Gets the parent directory. It achieves this by calling
            getDirectory() on the parentLookup directory passed into the
            constructor. If the other constructor which does not supply this
            parameter then getDirectory() returns null and an error condition
            is set (call lastError() to get this value)

          The parameter supplied to getDirectory() is the value passed in to the
          constructor has fullPathName with the last segment removed.

          \returns The parent directory or null on error or if there is no
                   parent.
         */
        Directory *getParent();

        /** Looks up the supplied pathname to try and find the associated
          directory. It only searches for directories below this one.
          \param directoryName The full directory name using forward (/)
          slashes.
         */
        Directory *getDirectory(QString directoryName);

        /** Refreshes the directory.
         * This reads the entire directory in again from disk. There shouldn't
         * be any need to call this after the object has been created unless
         * something else has changed the on-disk structures.
         */
        void refresh();

        /** Gets the directories full pathname.
          \returns The directories full path name (starting with /)
         */
        QString getFullPathName() {return fullPathName; }

        /** Gets the disks volume label. FAT sometimes stores the disks volume
            label as a fake directory entry in the root directory. This
            function retrieves that value. As such, it will only work if the
            directory object is the Root Directory. For subdirectories it will
            just return an empty string.

          \returns A volume label (if this is the root directory and one has
                   been set there) or an empty string (if this is not the root
                   directory or none has been set there).
         */
        QString getVolumeLabel() {return volumeLabel; }

    private:

        QString fullPathName;   /*!< The full path name of this directory */
        QString parentPathName; /*!< The path name of the parent
                                     (value passed in) */
        QString volumeLabel;    /*!< The disks volume label. Only set for the
                                     root directory. */
        QPointer<FatFS> filesystem; /*!< The filesystem we are a part of */
        QList<DirectoryEntry *> dirents; /*!< All directory entries */


        /** Finds a subdirectory somewhere under this level.
          \param directoryName the name of the directory to look for.
          \returns The Directory object if it could be found, otherwise NULL.
         */
        Directory *findSubdirectory(QString directoryName);

        /** Requests the specified directory from the FatFS instance and
            returns it. It updates the Directory objects error code with any
            related errors that occur in the FatFS object while getting the
            directory.
          \param directoryName the directory to fetch
          \returns A new Directory object matching the supplied pathname or
                    NULL if unsuccessfull.
         */
        Directory *getDirFromFS(QString directoryName);

        /** Clears the dirents list.
         */
        void clearDirEntries();
    };
}

#endif // DIRECTORY_H
