/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "directoryentry.h"
#include "util.h"
#include <QtDebug>
#include <QStringList>
#include <QtAlgorithms>

using namespace FilesystemFAT;
using namespace FilesystemFAT::Errors;

DirectoryEntry::DirectoryEntry(QObject *parent) :
    QObject(parent)
{
    qDebug() << "Directory Entry constructor (ROOT DIRECTORY)";
    loadAttributes('\0');
    mEntryCluster = 0;
    mFileSize = 0;
}

DirectoryEntry::DirectoryEntry(QByteArray data,
                               QObject *parent) :
    QObject(parent)
{
    initialise(data);
}

DirectoryEntry::~DirectoryEntry() {
    // Clean up all long filename components that we own.
    for (int i = 0; i < lfnComponents.count(); i++)
        delete lfnComponents.takeFirst();
}

QString DirectoryEntry::fileName() const {
    if (mLongFilename.isEmpty())
        return shortFilename();

    return mLongFilename;
}

QString DirectoryEntry::shortFilename() const {
    // If its a volume label then it shouldn't have a '.'
    if (!isVolumeLabel())
        return shortBaseName().trimmed() +
         (!shortExtension().trimmed().isEmpty() ? "." + shortExtension().trimmed() : "");
    else
        return shortBaseName().trimmed() + shortExtension().trimmed();
}

QString DirectoryEntry::baseName() const {
    if (!mLongFilename.isEmpty()) {
        QStringList parts = mLongFilename.split(".",QString::KeepEmptyParts);
        parts.takeLast();
        return parts.join(".");
    } else {
        return shortBaseName();
    }
}

QString DirectoryEntry::shortBaseName() const {
    return mFilename;
}

QString DirectoryEntry::extension() const {

    if (!mLongFilename.isEmpty()) {
        QStringList parts = mLongFilename.split(".",QString::SkipEmptyParts);
        return parts.last();
    } else {
        return shortExtension();
    }
}

QString DirectoryEntry::shortExtension() const {
    return mExtension;
}

ReturnCode DirectoryEntry::initialise(QByteArray data) {

    // Make sure we have 32 bytes.
    if (data.length() > 32) {
        lastErr = Errors::TOO_MUCH_DATA;
        return lastErr;
    } else if (data.length() < 32) {
        lastErr = Errors::NOT_ENOUGH_DATA;
        return lastErr;
    }

    /* Ok, We need to do the following:
       1) Split up the input data. In order of appearance, this is:
            FileName, Attributes, Reserved, EA Data, date/time stamp,
            Entry Cluster, File Size
       2) Then we need to convert strings, etc to the right format
       3) Then we need to decode the MS-DOS time/date stamp
       4) Lastly, we need to decode the attributes.
     */

    // Store the data for future use
    entryData = data;

    // The FileName - bytes 0-7
    QByteArray t_filename;
    for (int i = 0; i <= 7; i++) {
        t_filename.append(data.at(i));
    }

    // The file extension - 3 characters, 3 bytes from 8 to 10
    QByteArray t_extension;
    for (int i = 8; i <= 10; i++) {
        t_extension.append(data.at(i));
    }

    // Attributes - byte 11
    char t_attributes = data.at(11);


    // Reserved - two bits are used by NT and later versions to encode case
    // information. Otherwise, it should be 0.
    //char c_res_1 = data.at(12);

    // Created time, fine resolution - 10ms units from 0 to 199
    //char ct_ms = data.at(13);

    // Created time - hour, minute and second.
    QByteArray ct_time;
    ct_time.append(data.at(14));
    ct_time.append(data.at(15));
    mCreatedTime = convertTime(data.at(14),data.at(15));

    // Created date - year, month and day
    QByteArray ct_date;
    ct_date.append(data.at(16));
    ct_date.append(data.at(17));
    mCreatedDate = convertDate(data.at(16),data.at(17));

    // Last access date - year, month and day
    QByteArray at_date;
    at_date.append(data.at(18));
    at_date.append(data.at(19));
    mAccessDate = convertDate(data.at(18), data.at(19));

    // Extended Attribute Index (used by OS/2 and NT). In FAT32 this is the
    // high 2 bytes of the first cluster number.
    QByteArray ea_data;
    ea_data.append(data.at(20));
    ea_data.append(data.at(21));

    // Last Modified time.
    QByteArray mt_time;
    mt_time.append(data.at(22));
    mt_time.append(data.at(23));
    mModifiedTime = convertTime(data.at(22),data.at(23));

    // Last Modified date.
    QByteArray mt_date;
    mt_date.append(data.at(24));
    mt_date.append(data.at(25));
    mModifiedDate = convertDate(data.at(24),data.at(25));

    // Entry cluster - bytes 26 and 27.
    // Now, this is a 32bit integer stored little-endian.
    QByteArray t_entry_cluster;
    t_entry_cluster.append(data.at(26));
    t_entry_cluster.append(data.at(27));
    t_entry_cluster.append('\0');
    t_entry_cluster.append('\0');

    // File size - bytes 28 to 31
    QByteArray t_file_size;
    for (int i = 28; i <= 31; i++) {
        t_file_size.append(data.at(i));
    }

    // Ok, we have all the data. Now time to convert it!
    mFilename = QString(t_filename);
    mExtension = QString(t_extension);
    mFirstFNChar = t_filename.at(0);

    loadAttributes(t_attributes);

    // Get the integers
    mEntryCluster = Util::byteArrayToIntS(t_entry_cluster);
    mFileSize = Util::byteArrayToIntS(t_file_size);

    // Compute the MS-DOS Filename Checksum
    fnChecksum = 0;
    QString shortName = mFilename + mExtension;
    for (int i = 0; i < shortName.length(); i++)
        fnChecksum = ((fnChecksum & 1) << 7) +
                     (fnChecksum >> 1) + shortName.at(i).toAscii();

    return SUCCESS;
}

void DirectoryEntry::loadAttributes(char attributes) {
    mAttributes = attributes;

    maReadOnly = Util::getBit(attributes,0);
    maHidden = Util::getBit(attributes,1);
    maSystem = Util::getBit(attributes,2);
    maVolumeLabel = Util::getBit(attributes,3);
    maSubdirectory = Util::getBit(attributes,4);
    maArchive = Util::getBit(attributes,5);
    maDevice = Util::getBit(attributes,6);
    maUnused = Util::getBit(attributes,7);
}

QTime DirectoryEntry::convertTime(char byteA, char byteB) {
    /* MS-DOS stores the time as follows:
         Bits      Field        Range
         0-4     Seconds/2      0-29
         5-10     Minutes       0-59
         15-11     Hours        0-23
     */

    // First up, lets get both bytes into one integer so we can move and mask
    // stuff freely.
    QByteArray qba;
    qba.append(byteA);
    qba.append(byteB);
    qba.append('\0');
    qba.append('\0');

    int data = Util::byteArrayToIntS(qba);

    // Ok, the two bytes should be at the start of the int data.

    // Now, we are after the Seconds field. This is the first 5 bits. This
    // means we need to mask everything else out.
    int mask = 0x0000001F;

    // Now we mask out the unwanted data
    int seconds = data & mask;

    // Multiply by 2 (as DOS only records every second second
    seconds = seconds * 2;
    // And we now have the number of seconds.

    // Next up, the Minutes field. First we must shift the data to the right
    // by 5 bits:
    data = data >> 5;

    // Now we want the lower 6 bits - thats 0x3F
    mask = 0x0000003F;

    // Mask out the unwanted stuff (the hours)
    int minutes = data & mask;
    // And we should have the minutes now.

    // Lastly, the Hours field. All we have to do for this is shift. No masking
    // required.
    int hours = data >> 6;

    // And now we just have to convert to QTime and we are done!
    QTime t(hours,minutes,seconds,0);

    return t;
}

QDate DirectoryEntry::convertDate(char byteA, char byteB) {
    /* MS-DOS stores the date as follows:
         Bits      Field        Range   Comments
         0-4        Day         1-31
         5-8       Month        1-12    1=January, 12=December
         9-15      Year         0-127   0-1980, 127=2107
     */

    // Firstly, put all the data in an int
    QByteArray qba;
    qba.append(byteA);
    qba.append(byteB);
    qba.append('\0');
    qba.append('\0');

    int data = Util::byteArrayToIntS(qba);

    // Now make a mask for the lower 5 bits
    int mask = 0x0000001F;

    // Pull out the day
    int day = data & mask;

    // Shift the day data out of the way
    data = data >> 5;

    // Prepare the mask to give us the month
    mask = 0x0000000F;

    // Pull out the month
    int month = data & mask;

    // Move the month out of the way to give us years
    int year = data >> 4;

    // Now, the years start at 1980 so we should just add
    // 1980 to the year we have so far:
    year += 1980;

    // And then convert it all to a QDate
    QDate d(year,month,day);

    return d;
}

/** For checking if one LfnComponents sequence number is lower than anothers.
  \param c1 The first LfnComponent.
  \param c2 The second LfnComponent.
  \returns If c1 is less than c2.
 */
bool lfnComponentLessThan(const LfnComponent *c1, const LfnComponent *c2)
{
    return c1->getSequenceNumber() < c2->getSequenceNumber();
}

void DirectoryEntry::addLFNComponent(LfnComponent *component) {

    // Check that the component matches this file before accepting it.
    if (component->getDOSFilenameChecksum() != getFilenameChecksum()) {
        qDebug() << "DirectoryEntry discarding passed LfnComponent due to "
                    "checksum mismatch.";
        delete component;
    }

    // Take ownership of the component now that we have accepted it.
    lfnComponents.append(component);
    component->setParent(this);

    // Sort the component list
    qSort(lfnComponents.begin(),lfnComponents.end(),lfnComponentLessThan);

    mLongFilename = "";
    for (int i = 0; i < lfnComponents.count(); i++) {
        mLongFilename += lfnComponents.at(i)->getFilenameData();
    }
    mLongFilename = mLongFilename.trimmed();
}
