/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef DIRECTORYENTRY_H
#define DIRECTORYENTRY_H

#include <QObject>
#include <QByteArray>
#include <QTime>
#include <QDate>
#include <QDateTime>
#include <QList>

#include "errors.h"
#include "lfncomponent.h"

namespace FilesystemFAT {

    class Directory;

    class DirectoryEntry;

    /** A shared pointer to a DirectoryEntry.
     */
    typedef QSharedPointer<DirectoryEntry> DirEntPtr;

    /** This class represents an entry in a directory. It has functions
      for decoding the 32 byte directory entries and providing access to
      its contents. It is intended for use by entries representing actual real
      files and directories. For dealing with Long Filename segments, see
      LfnComponent.
     */
    class DirectoryEntry : public QObject
    {
    Q_OBJECT
    public:
        /** Constructs a new directory entry.
          \param parent The parent QObject.
         */
        explicit DirectoryEntry(QObject *parent = 0);

        /** Constructs and initialises a new directory entry. It is
          the same as using the other constructor and then calling
          initialise(). You should check lastError() after using this.

          The supplied data must be 32 bytes long - any more or less
          and you'll get an error.

          \param data The data to initialise the directory entry from
          \param parent The parent QObject
         */
        explicit DirectoryEntry(QByteArray data,
                                QObject *parent = 0);

        /** Cleans up the directory entry.
          */
        ~DirectoryEntry();

        /** Initialises the directory entry using the supplied data. The
            supplied data must be 32 bytes long - any more or less and you'll
            get an error.
          \param data The data to initialise the directory entry with.
          \returns An error code.
         */
        Errors::ReturnCode initialise(QByteArray data);

        /** Returns the last error code to occur.
          \returns The last error code to occur.
         */
        Errors::ReturnCode lastError() const {return lastErr; }


        /** Gets the filename. If a long filename is present then that will be
            returned instead of the short filename. To always get the short
            filename use shortFilename().
          \returns The Filename.
         */
        QString fileName() const;

        /** Returns the entries short filename. If there is no long filename
            present this function will return the same value as fileName().
          \returns The short filename.
         */
        QString shortFilename() const;

        /** Returns the filename without extension. If a long filename is
         *  present then the basename returned by this function will be taken
         * from that. To always get the short basename, use the shortBaseName()
         * function.
         * \returns The filename without extension.
         */
        QString baseName() const;

        /** Always returns the short basename even if a long filename is
            present.
          \returns the short base name.
         */
        QString shortBaseName() const;

        /** Gets the files extension. If a long filename is present, the
            extension will be taken from that. To always get the extension of
            the short filename, use shortExtension().
          \returns The files extension
          */
        QString extension() const;

        /** Always returns the extension component of the short filename.
          \returns The extension component of the short filename.
         */
        QString shortExtension() const;

        /** Gets the files entry cluster. This is not much use outside of the
          FAT Filesystem code
          \returns The files entry cluster.
          */
        unsigned int entryCluster() const {return mEntryCluster; }

        /** Returns the files size. This should be zero if isSubdirectory()
            returns true.
          \returns The files size.
         */
        unsigned int size() const {return mFileSize; }

        /** Returns True if the Read Only attribute is set.
          \returns The value of the Read Only attribute.
         */
        bool isReadOnly() const {return maReadOnly; }

        /** Returns True if the Hidden attribute is set.
          \returns The value of the Hidden attribute.
         */
        bool isHidden() const {return maHidden; }

        /** Returns True if the System attribute is set.
          \returns The value of the System attribute.
         */
        bool isSystem() const {return maSystem; }

        /** Returns True if the Volume Label attribute is set. This
          indicates that this directory entry isn't a file but rather
          stores the volume label.
          \returns The value of the Volume Label attribute.
         */
        bool isVolumeLabel() const {return maVolumeLabel; }

        /** Returns True if the Subdirectory attribute is set.
          \returns The value of the Subdirectory attribute.
         */
        bool isSubdirectory() const {return maSubdirectory; }

        /** Returns True if the Archive attribute is set.
          \returns The value of the Archive attribute.
         */
        bool isArchive() const {return maArchive; }

        /** Returns True if the Device attribute is set. This attribute
          is only used internally by MS-DOS and should never be seen on-disk.
          \returns The value of the device attribute.
         */
        bool isDevice() const {return maDevice; }

        /** Gets the time and date the file was created.
          \returns The time and date the file was created.
         */
        QDateTime created() const {
            return QDateTime(mCreatedDate, mCreatedTime);
        }

        /** Gets the date the file was last accessed.
          \returns The date the file was last accessed.
         */
        QDate accessDate() const {return mAccessDate; }

        /** Gets the time and date when the file was last modified.
          \returns teh time and date the file was last modified.
         */
        QDateTime modified() const {
            return QDateTime(mModifiedDate, mModifiedTime);
        }

        /** Returns the entry data used to construct this directory entry.
          Only useful inside the Filesystem Implementation.
          \returns The entry data with which this DirectoryEntry was
            constructed.
         */
        QByteArray getEntryData() const {return entryData;}

        /** Returns true if the directory entry represents a file that has
          been deleted.
          \returns True if the file has been deleted.
         */
        bool isDeleted() const {return (mFirstFNChar == 0xE5); }

        /** Returns true if the directory entry is free (available for re-use).
          This includes entries representing deleted files.
          \returns True if the directory entry is available for reuse.
         */
        bool isFree() const {return ((mFirstFNChar == 0xE5) || (mFirstFNChar == 0x00)); }

        /** Returns true if the entry is available and no subsequent entry is
          in use.
          \returns True if this and all subsequent entries are unusued.
         */
        bool isFinalEntry() const {return (mFirstFNChar == 0x00); }

        /** Returns true if the entry is a dot entry (. for current directory
          or .. for parent directory).
          \returns True if the filename starts with the '.' character.
         */
        bool isDotEntry() const {return (mFirstFNChar == '.');}

        /** So that Directories can call private functions to add
          long filename components and access the MS-DOS filename checksum.
         */
        friend class Directory;

    protected:
        Errors::ReturnCode lastErr;  /*!< The last error to occur */
        QByteArray entryData;        /*!< The data this directory entry was
                                          loaded from */

        // File Data
        QString mFilename;           /*!< The filename */
        QString mLongFilename;       /*!< The long filename (if present) */
        QString mExtension;          /*!< The extension */
        unsigned char mFirstFNChar;  /*!< The first character in the filename */
        unsigned char mAttributes;   /*!< The Attributes */
        unsigned int mEntryCluster;  /*!< The Entry Cluster */
        unsigned int mFileSize;      /*!< The File Size */
        unsigned char fnChecksum;    /*!< The MS-DOS Filename checksum */
        QList<LfnComponent *> lfnComponents; /*!< Long filename components */

        // Attributes
        bool maReadOnly;            /*!< Read Only attribute */
        bool maHidden;              /*!< Hidden attribute */
        bool maSystem;              /*!< System attribute */
        bool maVolumeLabel;         /*!< Volume Label attribute */
        bool maSubdirectory;        /*!< Subdirectory attribute */
        bool maArchive;             /*!< Archive attribute */
        bool maDevice;              /*!< Device attribute */
        bool maUnused;              /*!< Final unused bit */

        // Dates and Times
        QTime mCreatedTime;          /*!< Time of Creation */
        QDate mCreatedDate;          /*!< Date of Creation */
        QDate mAccessDate;           /*!< Date of last access */
        QTime mModifiedTime;         /*!< Time of last modification */
        QDate mModifiedDate;         /*!< Date of last modification */

        /** Loads the attribute values from the supplied char and populates the
          private member variables.
          \param attributes The attributes byte.
          */
        void loadAttributes(char attributes);

        /** Reads the MS-DOS Timestamp out of the two supplied bytes and
            converts it to a QTime.

          \param byteA The first byte of the MS-DOS Timestamp
          \param byteB The second byte of the MS-DOS Timestamp
          \returns The MS-DOS Timestamp converted to a QTime.
         */
        QTime convertTime(char byteA, char byteB);

        /** Reads the MS-DOS Datestamp out of the two supplied bytes and
            converts it to a QDate

          \param byteA The first byte of the date
          \param byteB The second byte of the date
          \returns The MS-DOS Date converted to a QDate.
         */
        QDate convertDate(char byteA, char byteB);

        /** Gets the MS-DOS Filename Checksum. Used for matching long filename
          components up to directory entries.
          \returns The MS-DOS Filenames checksum.
         */
        unsigned int getFilenameChecksum() const {
            return (unsigned int)fnChecksum;
        }

        /** Adds a long filename component. This function takes ownership of
            anything passed into it.

            If the components checksum matches the checksum of this directory
            entry, it will be added to an internal list and have its parent set
            to this directory entry.

            If the components checksum does not match it will be deleted and
            the function will return.

          \param component The long filename component.
         */
        void addLFNComponent(LfnComponent *component);
    };
}

#endif // DIRECTORYENTRY_H
