/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#include "fat12.h"
#include <QtDebug>

using namespace FilesystemFAT;

Fat12::Fat12(QIODevice *device,
             BiosParamBlock *bpb,
             unsigned int fatNumber,
             QObject *parent) :
    AbstractFAT(device,bpb,parent)
{
    this->fatNumber = fatNumber;

    init();
}

Fat12::Fat12(QIODevice *device,
             BiosParamBlock *bpb,
             QObject *parent) :
    AbstractFAT(device,bpb,parent)
{
    this->fatNumber = 0;

    init();
}

void Fat12::init() {
    // Locate the desired FAT and figure out its size
    const unsigned int reservedBytes = bpb->reservedSectors() *
                                       bpb->bytesPerSector();
    const unsigned int fatSize = bpb->sectorsPerFAT() * bpb->bytesPerSector();
    const unsigned int fatLocation = reservedBytes + (fatNumber * fatSize);

    /* Do a bit of sanity checking on our computed values */
    if (fatLocation > dev->size()) {
        lastErr = Errors::INV_FAT_LOC;
        return;
    }

    if (fatSize > dev->size()) {
        lastErr = Errors::INV_FAT_SIZE;
        return;
    }

    if (!dev->seek(fatLocation)) {
        lastErr = Errors::SEEK_ERR;
        return;
    }

    // Load the FAT
    fatData = dev->read(fatSize);

    if (fatData.count() < (int)fatSize) {
        qWarning() << "Failed to load FAT data.";
        lastErr = Errors::FAILURE;
        return;
    }

    lastErr = Errors::SUCCESS;
}

QList<unsigned int> Fat12::getChain(unsigned int startCluster) {
    QList<unsigned int> chain;

    /* Here we need to get a chain of clusters. Each value we get points
     * to the next value. The final value is 0xFFF which is where we stop.
     * This value isnt added to the list.
     */
    int current_value = startCluster;
    while (current_value < 0xFF8) {

        // Make sure the specified cluster isn't already in our list. If it is
        // then this part of the FAT is corrupt and we have an infinite loop.
        // For obvious performance reasons, a FAT-16 or FAT-32 implementation
        // of this class should probably use a QSet here for keeping track of
        // already seen values instead of looking them up in a QList.
        if (chain.contains(current_value)) {
            qWarning() << "Loop detected in file allocation table. Failing.";
            lastErr = Errors::LOOP_DETECTED;
            return chain;
        }


        // Make sure the value coming from the FAT is valid before we append it
        // and ask for the next.
        if (current_value == 0x001) {
            // 0x001 is a reserved value

            qWarning() << "Encountered reserved value 0x001 in FAT";
            lastErr = Errors::FAT_RES_A;
        } else if (current_value >= 0xFF0 && current_value <= 0xFF6) {
            // values between 0xFF0 and 0xFF6 are reserved.

            qWarning() << "Encountered reserved value between 0xFF0 and 0xFF6 "
                    "in FAT";
            lastErr = Errors::FAT_RES_B;
        } else if (current_value == 0xFF7) {
            // 0xFF7 is marked as a bad cluster. It shouldn't be a part of any
            // files cluster chain and has no next cluster.

            qWarning() << "Encountered cluster marked as bad in FAT.";
            lastErr = Errors::FAT_CLUST_BAD;
        }

        // If we found an error before, don't continue.
        if (lastErr != Errors::SUCCESS) break;

        // add the value
        chain.append(current_value);

        // Then get the next one
        current_value = getValue(current_value);

        // Abort if we hit an error.
        if (lastErr != Errors::SUCCESS) break;
    }

    return chain;
}

unsigned int Fat12::getValue(unsigned int clusterNumber) {

    if ((int)(clusterNumber * 1.5) > fatData.length()) {
        qWarning() << "Invalid cluster number. Returning termination value.";
        lastErr = Errors::INVALID_CLUSTER;
        return 0xFFF;
    }

    /* This code originally came from the libfat 0.3 file libfat.c. The
       specific function was:
           static int fat12_read_entry(Volume_t *V,
                                       DWORD N,
                                       int FatNum,
                                       DWORD *Value)
       which is GPL licensed and copyright:
           Copyright (C) 2006-2007  Paolo Angelelli <angelell@cs.unibo.it>
     */

    unsigned short val;
    if (clusterNumber%2 == 0) {
        val = 0;

        ((unsigned char *) &val)[0] = (unsigned char) fatData.at(
                                            (int) (clusterNumber * 1.5));
        ((unsigned char *) &val)[1] = (unsigned char) fatData.at(
                                            ((int) (clusterNumber * 1.5)) + 1);

        ((unsigned char *) &val)[1] &= 0x0F;
    } else {
        val =0;

        ((unsigned char *) &val)[0] = (unsigned char) fatData.at(
                                            ((int) (clusterNumber * 1.5)));
        ((unsigned char *) &val)[1] = (unsigned char) fatData.at(
                                            ((int) (clusterNumber * 1.5)) + 1);

        ((unsigned char *) &val)[0] = (((unsigned char *) &val)[0] >> 4);
        ((unsigned char *) &val)[0] |= (((unsigned char *) &val)[1] << 4);
        ((unsigned char *) &val)[1] = (((unsigned char *) &val)[1] >> 4);
    }

    lastErr = Errors::SUCCESS;
    return val;
}
