/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

/** \mainpage FatFS FAT Filesystem Implementation
 *
 * \section intro_sec Introduction
 *
 * FatFS is a simple implementation of Microsofts FAT filesystem as used by
 * MS-DOS, Windows and various other operating systems. It is implemented using
 * the Qt class library to keep code tidy and simple.
 *
 * This has been written for use by DIMG as its FAT Filesystem module to allow
 * for files to be extracted from FAT-formatted floppy disk images. The design
 * is such that it should be easy to reuse it elsewhere however. Its all quite
 * self-contained within the FilesystemFAT namespace and does not use any DIMG
 * interfaces/classes/etc.
 *
 * The implementation requires Qt version 4.6.0 or higher due to its use of
 * QScopedPointer in some places.
 *
 * \section limitations_sec Limitaitons
 * - Read-only (Read/write may come at a later date)
 * - FAT12 only (FAT16 and FAT32 will probably be added some day too)
 * - Targeted at reading floppy disks (no partition handling yet)
 */


#ifndef FATFS_H
#define FATFS_H

#include <QObject>
#include <QList>
#include <QIODevice>

#include "directoryentry.h"
#include "directory.h"
#include "fsfile.h"
#include "abstractfat.h"
#include "biosparamblock.h"
#include "dataarea.h"
#include "datachain.h"

/** The FilesystemFAT namespace contains all classes and functions related to
    the FAT Filesystem implementation.
 */
namespace FilesystemFAT {

    /** This is a read-only implementation of Microsofts FAT filesystem.
        Currently only the 12-bit version on Floppy Disks is supported.

        This FAT implementation supports long filenames which can be turned on
        or off using setLFNSupportEnabled().

        This class is single use - once you've opened the filesystem you can't
        close it. Deleting the instance will close the filesystem automatically.

        The FatFS class loads the filesystem and sets everything going. It
        provides various services to Directory and FsFile objects such as
        directory creation, etc. As such, once the FatFS object is deleted all
        associated FsFile and Directory objects will stop working properly.
        Attempts to use them after the FatFS object has been deleted may result
        in unpredictable behaviour and stability issues.

        Pointers to FsFiles and Directories are owned by the FatFS instance and
        will be automatically deleted when the FatFS instance is deleted.
     */
    class FatFS : public QObject
    {
        Q_OBJECT
    public:
        explicit FatFS(QObject *parent = 0);

        /** Opens the filesystem. Call lastError() to see what went wrong if
          open failed.
          \returns True if the open succeeded, false otherwise.
         */
        bool open();

        /** Returns true if the filesystem is currently open.
          \returns True if the filesystem is currently open.
         */
        bool isOpen();

        /** Creates a new directory object for the specified directory. If the
          directory does not exist, null is returned.

          \param directoryName The name of the directory to create.
          \returns A new Directory object.
         */
        Directory* createDirectory(QString directoryName);

        /** Creates a new file object for the specified file. If the file does
          not exist, null is returned.
          \param fileName The name of the file to create.
          \returns A new FsFile object or NULL.
         */
        FsFile* createFile(QString fileName);

        /** Returns the last error code to occur.
          \returns The last error code to occur.
         */
        Errors::ReturnCode lastError() {return lastErr; }

        void setDevice(QIODevice *device);

        /** Returns a pointer to the QIODevice being used by this filesystem
            instance.

            \returns A pointer to the QIODevice used by the filesystem.
         */
        QIODevice* getDevice();

        /** Gets the volume label. This comes from the BIOS Parameter Block if
          there is one stored there, otherwise it comes from the root directory.
          \returns The volume label.
         */
        QString getVolumeLabel();

        /** Returns true if Long Filename support is currently enabled.
          \returns True if long filename support is currently enabled.
         */
        bool isLFNSupportEnabled() {return lfnEnabled; }

        /** Enable or disable long filename support. This only affects new
          directories. Any already created directories will continue to use the
          previous value.
          \param enabled Enable long filename support or disable it.
         */
        void setLFNSupportEnabled(bool enabled) {lfnEnabled = enabled;}

        /** This is necessary so that Directory is able to get at the following
          functions:
            getFAT()
            getBPB()
            getDataArea()
            getDevice()
         */
        friend class Directory;

        /** This is necessary so that FsFile is able to get at the following
          functions:
            getFAT()
            getBPB()
            getDataArea()
            getDevice()
         */
        friend class FsFile;

        /** So that DataChain can get access to our file allocation tables,
          bios parameter block and data area.
         */
        friend class DataChain;

    private:
        /** Returns a pointer to the specified File Allocation Table. If the
          requested FAT does not exist the first one will be returned instead.
          This is owned by the FatFS object.
          \param number The FAT to get.
          \returns A pointer to the File Allocation Table.
         */
        AbstractFAT* getFAT(int number = 0);

        /** Returns a pointer to the BIOS Parameter Block. This is owned by the
          FatFS object.
          \returns A pointer to the BIOS Paramter Block.
         */
        BiosParamBlock* getBPB();

        /** Returns a pointer to the DataArea. This is used to get at file data
            in the filesystem. The pointer is owned by the FatFS instance.
            \returns A pointer to the Data Area.
         */
        DataArea* getDataArea();

        QList<AbstractFAT*> fats;   /*!< The File Allocation Table. */
        BiosParamBlock* bpb;        /*!< The BIOS Parameter Block */
        DataArea* da;               /*!< The data area of the filesystem
                                         (where files live) */
        QIODevice *dev;             /*!< The device we are accessing the
                                         filesystem from */
        Errors::ReturnCode lastErr; /*!< The last error to occur */
        bool fs_open;               /*!< Stores if the Filesystem is open or
                                         not. */
        bool lfnEnabled;            /*!< If Long Filenames are enabled. */
        Directory* root;            /*!< The root directory. Used to create
                                         other directories. */
    };
}

#endif // FATFS_H
