/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 09/07/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/


#include "lfncomponent.h"
#include "util.h"

#include <QByteArray>
#include <QTextCodec>
#include <QtDebug>

using namespace FilesystemFAT;

LfnComponent::LfnComponent(QByteArray data, QObject *parent) :
    QObject(parent)
{
    qDebug() << "-------------------------------";
    qDebug() << "Loading Long Filename component";

    /* From the ever-helpful Wikipedia:
    -------------- --------- -------------------------------------------
    Byte Offset    Length    Description
    -------------- --------- -------------------------------------------
    0x00            1        Sequence Number
    0x01            10       Name characters (five UTF-16 characters)
    0x0b            1        Attributes (always 0x0F)
    0x0c            1        Reserved (always 0x00)
    0x0d            1        Checksum of DOS file name
    0x0e            12       Name characters (six UTF-16 characters)
    0x1a            2        First cluster (always 0x0000)
    0x1c            4        Name characters (two UTF-16 characters)
    -------------- --------- -------------------------------------------
    */

    char seqNum = data.at(0);
    sequenceNumber = (unsigned int)seqNum;
    qDebug() << "Sequence Number:" << sequenceNumber;

    QByteArray fnChunkA = data.mid(1,10);

    unsigned char attr = data.at(0x0B);

    if (attr != 0x0F) {
        lastErr = Errors::INV_LFN_COMP;
        qWarning() << "Invalid Long Filename component. Expected 0x0F at "
                "offset 0x0B.";
        return;
    }

    if (data.at(0x0C) != 0x00) {
        qWarning() << "Expected 0x00 at offset 0x0C";
    }

    checksum = (unsigned char)data.at(0x0D);
    qDebug() << "MS-DOS Filename Checksum:" << (unsigned int)checksum;

    QByteArray fnChunkB = data.mid(0x0E,12);

    char fcA = data.at(0x1A);
    char fcB = data.at(0x1B);
    if (fcA != 0x00 || fcB != 0x00)
        qWarning() << "Expected 0x0000 at offset 0x1A.";

    QByteArray fnChunkC = data.mid(0x1C,4);

    QByteArray combined;
    combined.append(fnChunkA);
    combined.append(fnChunkB);
    combined.append(fnChunkC);

    deletedEntry = Util::getBit(seqNum,7);
    lastEntry = Util::getBit(seqNum,6);

    if (deletedEntry)
        qDebug() << "This entry is marked as deleted.";

    // If this is the last entry it needs to be trimmed.
    if (lastEntry) {
        qDebug() << "This is the last entry in the long filename sequence. "
                "Trimming...";

        if ((unsigned char)combined.at(combined.length() - 2) == 0x00) {
            qDebug() << "0x00 found.";
            // We have a pair of 0x00s that need removing.
            combined = combined.mid(0,combined.length()-2);
        } else if ((unsigned char)combined.at(combined.length() - 2) == 0xFF) {
            qDebug() << "0xFF found.";
            // We have 0xFF padding. Search backwards until we find a null.
            for (int i = combined.length()-2; i >= 0; i -= 2) {
                if ((unsigned char)combined.at(i) == 0x00) {
                    combined = combined.mid(0,i);
                    qDebug() << "0xFF Trimmed to" << i;
                    break;
                } else if ((unsigned char)combined.at(i) != 0xFF) {
                    // Oops. We couldn't find the null. We'll leave the
                    // string alone.
                    qWarning() << "Unexpected character attempting to trim "
                            "final filename segment.";
                    break;
                }
            }

        } else {
            // else the string isn't padded.
            qDebug() << "No padding detected. No trim necessary.";
        }
    }

    QTextCodec * codec = QTextCodec::codecForName("UTF-16");

    filenameData = codec->toUnicode(combined);
    qDebug() << "Filename Data:" << filenameData;

    qDebug() << "Long Filename Chunk loaded successfully.";
    qDebug() << "----------------------------------------";
    lastErr = Errors::SUCCESS;
}

QString LfnComponent::getFilenameData() const {
    if (!deletedEntry) return filenameData;
    return "";
}

bool LfnComponent::operator<(const LfnComponent &other) const {

    return (other.getSequenceNumber() < getSequenceNumber());
}
