/*
-----------------------------------------------------
   Application Name: "DIMG"
         Start Date: 27/09/2010
       Last Changed: $LastChangedDate: $
           Revision: $Revision: $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
-----------------------------------------------------
        This file is part of "DIMG".

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

        File Notes:
        ------------------
*/

#ifndef LFNCOMPONENT_H
#define LFNCOMPONENT_H

#include <QObject>
#include <QByteArray>

#include "errors.h"

namespace FilesystemFAT {

    /** This class represents a component in a long filename. It is constructed
      from a 32byte directory entry but does not inherit DirectoryEntry as
      most of the services it provides do not apply here.

      In the FAT Filesystem, long filenames are chopped up into 13 character
      chunks and stored in fake directory entries with the attributes set to
      0x0F so that MS-DOS will ignore them.

      Each fake directory entry contains 13 UTF-16 characters, a sequence number
      (the order in which the components should be arranged) and a checksum
      value for the actual files short filename. This checksum value is used to
      match up the LFN components to the actual real directory entries for the
      files.

      This class will check to make sure that the attributes (which live at
      offset 0x0B in the directory entry) equal 0x0F (Volume Label, System,
      Hidden, and Read Only). If the attribute byte is not 0x0F then the
      constructor will just set the Errors::INV_LFN_COMP error code and return.
     */
    class LfnComponent : public QObject
    {
    Q_OBJECT
    public:
        /** Constructs a new LfnComponent.
          \param data The data to construct from
          \param parent The parent QObject.
         */
        explicit LfnComponent(QByteArray data, QObject *parent = 0);

        /** Gets this components filename data. If bit 8 of the sequence number
            is set then an empty string will be returned (as bit 8 marks a
            deleted entry).
          \returns This components filename data.
         */
        QString getFilenameData() const;

        /** Gets the MS-DOS Filename Checksum.
          \returns The MS-DOS Filename cehcksum.
         */
        unsigned char getDOSFilenameChecksum() const {return checksum; }

        /** Gets the components sequence number (the order it should appear in
            the filename)
          \returns The components sequence number.
         */
        unsigned int getSequenceNumber() const {return sequenceNumber; }

        /** Returns True if this is the last component in the long filename
            sequence.
          \returns True if this is the last Long Filename component for the
                   file.
         */
        bool isLastEntry() const {return lastEntry;}

        /** Returns the last error to occur.
          \returns The code for the last error.
         */
        Errors::ReturnCode lastError() const {return lastErr; }

        /** Compares the sequence number of this object to other.
          \returns True if this object has a smaller sequence number.
         */
        bool operator<(const LfnComponent &other) const;
    private:
        QString filenameData;           /*!< The filename data */
        unsigned char checksum;         /*!< The DOS filename checksum */
        unsigned int sequenceNumber;    /*!< The component sequence number */
        Errors::ReturnCode lastErr;     /*!< The last error to occur */
        bool deletedEntry;              /*!< Entry is deleted (sequence number
                                             bit 8 is set) */
        bool lastEntry;                 /*!< This is the last entry (sequence
                                             number bit 7 is set) */
    };
}
#endif // LFNCOMPONENT_H
