/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: fatfslib.cpp
       Last Changed: $LastChangedDate: 2010-04-04 00:18:51 +1300 (Sun, 04 Apr 2010) $
           Revision: $Revision: 199 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2008,2009,2010
            License: GNU General Public License
   File Description: FatFS Filesystem Implementation
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/

#include <QFile>
#include <QtDebug>
#include "fatfslib.h"
#include "FatFS/util.h"
#include "FatFS/fsfile.h"
#include "fatentry.h"

bool FatFSlib::load(QIODevice * dev){
    // If there isnt a filesystem, make one
    if (fs == 0) fs = new FatFS(this);

    // Load the filesystem
    fs->setDevice(dev);
    fs->setLFNSupportEnabled(true);
    fs->open();
    is_closed = !fs->isOpen();
    return true;
}

bool FatFSlib::close(){
    if (fs == 0) return false;
    delete fs;
    fs = 0;
    is_closed = true;
    return true;
}

FSPlugin::Features FatFSlib::getFeatures() const{
    // At the moment we dont support Configure.

    FSPlugin::Features f;

    // First, the filesystem features:
    f |= FSPlugin::F_ReadOnly | FSPlugin::F_VolumeLabel;

    // Then the GUI features:
    f |= FSPlugin::F_DisplayAbout;

    // Then the file date/time features:
    f |= FSPlugin::F_CreationDate | FSPlugin::F_CreationTime;
    f |= FSPlugin::F_AccessDate;  // Access time isnt stored
    f |= FSPlugin::F_ModDate | FSPlugin::F_ModTime;

    // Then return it.
    return f;
}

FSPlugin::Attributes FatFSlib::getSupportedAttributes() const{
    /* FAT12 supports the ReadOnly, Hidden, System, VolumeID, Directory and
       Archive attributes.

       However, the VolumeID attribute should only be used internally by the
       filesystem - DOS never exposes it.
     */
    FSPlugin::Attributes a;

    a = a | A_ReadOnly | A_Hidden | A_System | A_Directory | A_Archive;

    return a;
}

QString FatFSlib::volumeLabel() const {
    if (fs == 0) return "";
    return fs->getVolumeLabel();
}

int FatFSlib::getLastErrorNumber() const{
    if (fs == 0) return 0;
    return fs->lastError();
}

QString FatFSlib::getLastErrorString() const{
    if (fs == 0) return "";
    return Errors::errorToString(fs->lastError());
}

bool FatFSlib::copyFile(QString source_filename, QString destination_filename){

    FilesystemFAT::FsFile *file = fs->createFile(source_filename);

    if (file->lastError() != Errors::SUCCESS) {
        if (file != NULL) delete file;
        return false;
    }

    QFile localFile(destination_filename);
    if (!localFile.open(QIODevice::WriteOnly)) {
        delete file;
        return false;
    }

    localFile.write(file->readAll());
    localFile.flush();
    localFile.close();


    delete file;

    return true;
}

QByteArray FatFSlib::readFile(QString filename){
    FilesystemFAT::FsFile *file = fs->createFile(filename);

    if (file->lastError() != Errors::SUCCESS) {
        if (file != NULL) delete file;
        return false;
    }

    return file->readAll();
}

FSEntry2 * FatFSlib::getEntryByName(QString filename){
    // Try to create it as a directory first.
    Directory * direct;
    direct = fs->createDirectory(filename);

    if (fs->lastError() == Errors::SUCCESS) {
        // Seems the file requested is actually a Directory. We'll return
        // the entry for that.

        DirEntPtr e(direct);

        // DirEntPtr is a typedef for QSharedPointer<DirectoryEntry>. Above we are
        // placing a pointer to a DirectoryEntry under the control of QSharedPointer
        // so we can't allow the original DirectoryEntry to have any parent
        // (otherwise it will be under the control of the parent too and that is
        // bad).
        e->setParent(0);

        return new FatEntry(e,direct->getParent()->getFullPathName());
    } else {
        // Just in case it partially succeeded we had better delete what ever
        // is left over
        if (direct != 0) delete direct;

        QStringList qsl = filename.split("/",QString::SkipEmptyParts);
        QString file = qsl.takeLast();
        QString dir = qsl.join("/");
        if (dir.isEmpty()) dir = "/";


        // Ok, we are looking for [file] in [dir]
        direct = fs->createDirectory(dir);
        if (fs->lastError() != Errors::SUCCESS) {
            // TODO: Update last error code
            return 0;
        }

        DirEntPtr e(direct->getFile(file));
        if (direct->lastError() != Errors::SUCCESS) {
            // TODO: Update last error code
            return 0;
        }

        e->setParent(0);

        // We only wanted direct so we could create the file.
        delete direct;

        return new FatEntry(e,dir);
    }
}

QList<FSEntry *> FatFSlib::getEntryList(QString directory) {
    QList<FSEntry2 *> v2Entries = getEntryList2(directory);

    QList<FSEntry *> v1Entries;

    foreach(FSEntry2 *ent, v2Entries) {
        v1Entries.append(ent);
    }
    return v1Entries;
}

void FatFSlib::displayAbout() const {
    about_dialog ad;
    ad.exec();
}

QList<FSEntry2 *> FatFSlib::getEntryList2(QString directory){
    // Get the directory listing.

    QScopedPointer<Directory> dir;
    dir.reset(fs->createDirectory(directory));
    if (fs->lastError() != Errors::SUCCESS)
        return QList<FSEntry2 *>();

    QList<DirEntPtr> el;

    qDebug() << "Getting subdirectories...";
    QList<DirectoryPtr> dirs = dir->getSubdirs();
    if (dir->lastError() != Errors::SUCCESS) {
        // TODO: Update local last error
        return QList<FSEntry2 *>();
    } else {
        qDebug() << "Appending subdirectories";
        foreach(DirectoryPtr d,dirs) {
            el.append(d);
        }
    }

    qDebug() << "Getting files...";
    QList<FsFilePtr> files = dir->getFiles();
    if (dir->lastError() != Errors::SUCCESS) {
        // TODO: Update local last error
        return QList<FSEntry2 *>();
    } else {
        qDebug() << "Appending files";
        foreach(FsFilePtr f,files) {
            el.append(f);
        }
    }

    // If the directory listing was empty, return an empty entry list.
    if (el.isEmpty()) return QList<FSEntry2 *>();

    // Now we need to convert all the entires
    QList<FSEntry2 *> entryList;
    while (!el.isEmpty()) {
        DirEntPtr e = el.takeFirst();
        entryList.append(new FatEntry(e,directory));
    }
    return entryList;
}

Q_EXPORT_PLUGIN2(fatfslib, FatFSlib)
