/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: rawimagefile.cpp
         Start Date: 27/02/2010
       Last Changed: $LastChangedDate: 2011-04-09 00:17:11 +1200 (Sat, 09 Apr 2011) $
           Revision: $Revision: 295 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
   File Description: RawImageFile implementation
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/


#include "rawimagefile.h"

RawImageFile::RawImageFile(QObject *parent) :
    ImageFile(parent)
{
    file = 0;
    buffer = 0;
    currentDevice = 0;
    cached = false;
    createdInCache = false;
}

RawImageFile::~RawImageFile() {
    close();
}

void RawImageFile::open(QString filename,
                        bool create,
                        bool cache,
                        long long expectedSize) {
    qDebug() << "FPraw: Open requested for file: " << filename;

    if (create) {
        qDebug() << "FPraw: Requested open mode is CREATE. Will open the "
                "file and TRUNCATE if it exists.";

        close();

        // Should we cache it?
        if (cache) {
            // the user asked. Is the expected size too great?

            bool ok;

            // Open the settings file
            QScopedPointer<QSettings> settings;
            QString filename;
            filename = QCoreApplication::applicationName() + ".ini";
            qDebug() << "Checking for configuration file " << filename;

            if (QFile::exists(filename)) {
                // Load settings from there
                settings.reset(new QSettings(filename, QSettings::IniFormat, 0));
            } else {
                settings.reset(new QSettings());
            }

            long long maxCacheSize = settings->value("/FileHandling/maxCacheSize",
                                                     5242880).toLongLong(&ok);

            if (!ok) maxCacheSize = 5242880;

            // If the file size is less than 5MB (or the user specified maximum),
            // load it all into memory.
            if (expectedSize < maxCacheSize) {
                buffer = new QBuffer();
                buffer->open(QIODevice::ReadWrite);
                buffer->buffer().resize(expectedSize);
                currentDevice = buffer;               
                connect(currentDevice,
                        SIGNAL(bytesWritten(qint64)),
                        this,
                        SLOT(bytesWritten(qint64)));
                this->filename = filename;

                cached = createdInCache = true;

                cleanFile();

                return;
            }
        }
        // Ok. So we arent caching it. We will open the on-disk file then.

        file = new QFile(filename);
        if (!file->open(QIODevice::Truncate | QIODevice::ReadWrite)) {
            qDebug() << "Failed to open file.";
        }
        currentDevice = file;
        connect(currentDevice,
                SIGNAL(bytesWritten(qint64)),
                this,
                SLOT(bytesWritten(qint64)));
        cached = createdInCache = false;

    } else {
        qDebug() << "FPraw: Requested open mode is EXISTING. Will open the "
                "file only if it exists. It will not be truncated.";
        openExisting(filename, cache);
    }


}

void RawImageFile::openExisting(QString filename,
                                bool cache) {
    qDebug() << "FPraw\\RawImageFile: Attempting to open existing file: " << filename;

    // Open the settings file
    QScopedPointer<QSettings> settings;
    QString settingsFile;
    settingsFile = QCoreApplication::applicationName() + ".ini";
    qDebug() << "Checking for configuration file " << settingsFile;

    if (QFile::exists(settingsFile)) {
        // Load settings from there
        settings.reset(new QSettings(settingsFile, QSettings::IniFormat, 0));
    } else {
        settings.reset(new QSettings());
    }
    qDebug() << "Settings Loaded.";
    close();

    file = new QFile(filename);
    this->filename = filename;

    if (!file->open(QIODevice::ReadWrite)) {
        // Something has gone wrong
        QMessageBox::critical(0,"status","Unable to open file");
        return;
    }
    qDebug() << "File open.";

    // The current device shall be the file we just opened.
    currentDevice = file;
    connect(currentDevice,
            SIGNAL(bytesWritten(qint64)),
            this,
            SLOT(bytesWritten(qint64)));

    // Now, should we cache it?
    if (cache) {
        qDebug() << "Caching enabled";
        // Figure out if the user has changed the max cache size.
        bool ok;
        long long maxCacheSize = settings->value("/FileHandling/maxCacheSize",
                                                 5242880).toLongLong(&ok);

        if (!ok) maxCacheSize = 5242880;

        // If the file size is less than 5MB (or the user specified maximum),
        // load it all into memory.
        if (file->size() < maxCacheSize) {
            qDebug() << "File size is within cache limit.";
            QProgressDialog qpd;
            qpd.setWindowTitle("Loading...");
            qpd.setMaximum(file->size());
            if (buffer) delete buffer;
            buffer = new QBuffer();
            buffer->open(QIODevice::ReadWrite);

            int blocks = file->size() / 512;

            for (int i = 0; i < blocks; i++) {
                if (qpd.wasCanceled()) {
                    file->close();
                    delete file;
                    file = 0;
                    delete buffer;
                    buffer = 0;
                    return;
                }
                //qDebug() << "Reading block " << i;
                file->seek(i*512);
                buffer->seek(i*512);
                buffer->write(file->read(512));
                qpd.setValue(i);
            }
            qpd.setValue(file->size());

            // We were only loading into cache. Those writes shouldn't be considered
            // as making it dirty.
            cleanFile();

            // The current device is now the cache buffer instead of the file.
            currentDevice = buffer;
            connect(currentDevice,
                    SIGNAL(bytesWritten(qint64)),
                    this,
                    SLOT(bytesWritten(qint64)));
            cached = true;
        }
    }
    // If we shouldn't cache it we'll leave the current device as the file.
}

QIODevice *RawImageFile::device() {
    return currentDevice;

}

bool RawImageFile::createCachedFile() {
    if (createdInCache) {
        file = new QFile(filename);

        if (!file->open(QIODevice::Truncate | QIODevice::ReadWrite)) {
            qDebug() << "Failed to open file.";
            return false;
        }
        else
            return true;
    }
    return true;
}

int RawImageFile::save() {

    // Call this in case the file was created in cache.
    createCachedFile();

    // Send buffer contents to disk if caching is on but only
    // if it is dirty. If it is not cached then there is nothing to save.
    // with caching disabled all writes go straight to disk.
    if (cached && dirty) {
        qDebug() << "Writing out cache...";
        file->write(buffer->buffer());
    }

    file->flush();

    cleanFile();
    return 0;
}

int RawImageFile::close() {
    qDebug() << "FPRaw/RawImageFile: Attempting to close file";

    if (file) {
        file->close();
        delete file;
        file = 0;
    }
    if (buffer) {
        delete buffer;
        buffer = 0;
    }
    currentDevice = 0;
    cached = false;

    return 0;
}

ImageFile::Features RawImageFile::getFeatures() const {
    return ImageFile::F_Cached | ImageFile::F_RawFormat | ImageFile::F_Floppy_Disk;
}

bool RawImageFile::isOpen() const {
    return currentDevice->isOpen();
}

void RawImageFile::bytesWritten(qint64 /*bytes*/) {
    dirtyFile();
}

void RawImageFile::cleanFile() {
    dirty = false;
    qDebug() << "FPRaw/RawImageFile is now clean.";
}

void RawImageFile::dirtyFile() {
    dirty = true;
    qDebug() << "FPRaw/RawImageFile is now dirty.";
}
