/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: zipimagefile.h
         Start Date: 28/03/2010
       Last Changed: $LastChangedDate: 2011-06-12 16:41:13 +1200 (Sun, 12 Jun 2011) $
           Revision: $Revision: 310 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
   File Description: ZipImageFile header
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/


#ifndef ZIPIMAGEFILE_H
#define ZIPIMAGEFILE_H

#include <imagefile.h>
#include <QFile>
#include <QBuffer>
#include <QMessageBox>
#include <QSettings>
#include <QScopedPointer>
#include <QProgressDialog>
#include <QtDebug>
#include <QCoreApplication>
#include <QFileInfo>

#include "ziputil.h"

class ZipImageFile : public ImageFile
{
Q_OBJECT
public:
    explicit ZipImageFile(QObject *parent = 0);
    ~ZipImageFile();

    /** Opens a zip image file, optionally creating it if it does not
      already exist.
     \param filename The file to open or create
     \param create If the file should be created if it does not exist
     */
    void open(QString filename,
              bool create);

    /** Opens an existing image file. It can optionally cache the entire image
      file in memory if desired.

      \param filename The file to open.
     */
    void openExisting(QString filename);

    /** Gets the name of the currently open image file.
      \returns The name of the currently open image file.
     */
    QString fileName() const { return filename; }

    /** A device which can be used for read/write access to the image file
      \returns A QIODevice which can be used to read from and write to the device. This
      pointer belongs to the ZipImageFile and must not be deleted.
     */
    QIODevice *device();

    /** Saves the image file if any changes have been made.
      \returns An error code.
     */
    int save();

    /** Closes the image without saving it. Call save() to save it first.
      \returns An error code.
     */
    int close();

    /** Gets the features supported by this image file.
      \returns The image files supported features.
      */
    ImageFile::Features getFeatures() const;

    /** Returns the images comment.
      \returns The images comment.
     */
    QString getComment() const { return comment; }

    /** Sets the images comment.
      \param comment The new comment.
      \returns The images comment.
     */
    void setComment(QString comment) {
        // The comment must be converted to CR+LF for WinImage compatibility.
        this->comment = comment.replace("\x0A","\x0D\x0A");
        dirty = true;
    }

    long maxCommentLength() {return 32767; }

    bool isOpen() const;

    bool isDirty() const {return dirty;}

    int errorCode;
signals:

private slots:
    void bytesWritten(qint64 /*bytes*/);

private:
    void cleanFile();
    void dirtyFile();

    QString filename;
    QString imageName;
    QString comment;

    QBuffer *buffer;
    QIODevice *currentDevice;

    bool dirty;
};

#endif // ZIPIMAGEFILE_H
