/*
-----------------------------------------------------
   Application Name: DIMG
          File Name: ziputil.cpp
         Start Date: 28/03/2010
       Last Changed: $LastChangedDate: 2011-06-12 14:55:07 +1200 (Sun, 12 Jun 2011) $
           Revision: $Revision: 304 $
          Author(s): David Goodwin
          Copyright: (C) Copyright David Goodwin, 2010
            License: GNU General Public License
   File Description: various zip-related utility functions
-----------------------------------------------------
        This file is part of DIMG.

    This is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this software; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

File Notes:
------------------
*/

#include <QtDebug>
#include <QFile>

#define BUILD_DLL

#include "ziputil.h"

// Minizip headers
#include "../../lib/minizip/zip.h"
#include "../../lib/minizip/unzip.h"

/** Opens the specified zip file for READING.
  \param filename The zip file to open for reading.
  \returns A handle to the opened zipfile or NULL.
 */
unzFile openUnzipFile(QString filename) {
    unzFile uf       = NULL;

    /*
     * Open the zip file
     */
    qDebug() << "Attempting to open zip file" << filename << "for extraction";
    if (!filename.isEmpty())
        uf = unzOpen(filename.toAscii().constData());

    if (uf==NULL)
        qDebug() << "failed to open zip file.";
    else
        qDebug() << "Zip file open.";
    return uf;
}

/** Opens the specified zip file for WRITING.
  \param filename The zip file to open
  \param append How to open it
  \returns A zip file handle.
 */
zipFile openZipFile(QString filename, int append) {
    zipFile zf;
    zf = zipOpen(filename.toAscii().constData(), append);

    if (zf == NULL)
        qDebug() << "Error opening zip file" << filename;

    return zf;
}

ZUEXPORT int writeFileToZip(QString zipFileName,
                   QIODevice *fileData,
                   QString fileName,
                   QDateTime fileDateTime,
                   QString fileComment,
                   bool overwriteZipFileIfExists,
                   bool appendZipFileIfExists,
                   bool compressFileData,
                   int compressionLevel,
                   QProgressDialog *qpd) {

    // Figure out how we actually want to open this zip file.
    int appendValue = APPEND_STATUS_ADDINZIP;
    if (appendZipFileIfExists) {
        appendValue = APPEND_STATUS_CREATEAFTER;
    } else if (overwriteZipFileIfExists) {
        appendValue = APPEND_STATUS_CREATE;
    }
    if (!QFile::exists(zipFileName)) appendValue = APPEND_STATUS_CREATE;

    zipFile zf = openZipFile(zipFileName,appendValue);

    if (zf == NULL) {
        qWarning() << "Failed to open zip file.";
        return ZIP_ERRNO;
    }

    int err = ZIP_OK;
    int bufferSize = WRITEBUFFERSIZE;
    char* buffer = NULL;
    int sizeRead = 0;

    // We need RAM!!!
    buffer = (char*)malloc(bufferSize);
    if (buffer == NULL) {
        qWarning() << "Error allocating memory for write buffer.";
        zipClose(zf,NULL);
        return ZIP_INTERNALERROR;
    }

    // Setup the file info
    zip_fileinfo zi;
    zi.tmz_date.tm_sec = zi.tmz_date.tm_min = zi.tmz_date.tm_hour =
    zi.tmz_date.tm_mday = zi.tmz_date.tm_mon = zi.tmz_date.tm_year = 0;
    zi.dosDate = 0;
    zi.internal_fa = 0;
    zi.external_fa = 0;

    // Set the date/time. The DOS date time will be left empty for now.
    QTime time = fileDateTime.time();
    QDate date = fileDateTime.date();
    zi.tmz_date.tm_sec = time.second();
    zi.tmz_date.tm_min = time.minute();
    zi.tmz_date.tm_hour = time.hour();
    zi.tmz_date.tm_mday = date.day();
    zi.tmz_date.tm_mon = date.month();
    zi.tmz_date.tm_year = date.year();

    // Open a new file in the archive.
    err = zipOpenNewFileInZip2(zf,
                               fileName.toAscii().constData(),
                               &zi,
                               NULL,
                               0,
                               NULL,
                               0,
                               fileComment.isEmpty() ? NULL : fileComment.toAscii().constData(),
                               (compressFileData) ? Z_DEFLATED : 0,
                               compressionLevel,
                               0);
    if (err != ZIP_OK) {
        qWarning() << "Error opening new file in zip archive. Error code was" << err;
    } else {
        do {
            err = ZIP_OK;
            sizeRead = (int)fileData->read(buffer,bufferSize);

            if (sizeRead < bufferSize) {
                if (!fileData->atEnd()) {
                    qWarning() << "Error reading file data. Error was " << fileData->errorString();
                    err = ZIP_ERRNO;
                }
            }

            if (sizeRead > 0) {
                err = zipWriteInFileInZip(zf,(void*)buffer,sizeRead);
                if (err < 0) {
                    qWarning() << "Error writing data to zip file. Error code was" << err;
                }
            }

        } while ((err == ZIP_OK) && (sizeRead > 0));
    }

    if (err < 0)
        err = ZIP_ERRNO;
    else {
        err = zipCloseFileInZip(zf);
        if (err != ZIP_OK) {
            qWarning() << "Error closing new file in zip archive. Error code was" << err;
        }
    }

    err = zipClose(zf,NULL);
    if (err != ZIP_OK)
        qWarning() << "Error closing zip file. Error code was" << err;

    free(buffer);

    return err;
}

ZUEXPORT int setZipFileComment(QString zipFileName, QString comment) {

    int appendValue = APPEND_STATUS_ADDINZIP;

    if (!QFile::exists(zipFileName)) appendValue = APPEND_STATUS_CREATE;

    zipFile zf = openZipFile(zipFileName,appendValue);

    const char *globalComment = (comment.isEmpty()) ? NULL : comment.toAscii().constData();

    int err = zipClose(zf,globalComment);

    if (err != ZIP_OK)
        qDebug() << "Error closing zip file " << zipFileName;

    return err;
}

ZUEXPORT QString getZipFileComment(QString zipFileName) {
    unzFile uf = openUnzipFile(zipFileName);
    unz_global_info gi;
    int err;

    if (uf == NULL) return "";

    err = unzGetGlobalInfo(uf,&gi);

    if (err != UNZ_OK) {
        qWarning() << "Error in unzGetGlobalInfo. Error code was " << err;
        unzClose(uf);
        return "";
    }

    if (gi.size_comment == 0) {
        unzClose(uf);
        return "";
    }

    char *globalComment = new char[gi.size_comment+1];

    unzGetGlobalComment(uf,globalComment,gi.size_comment);

    // unzGetGlobalComment doesn't seem to null-terminate the buffer.
    globalComment[gi.size_comment] = '\0';

    QString comment = globalComment;

    delete globalComment;
    unzClose(uf);

    return comment;
}

ZUEXPORT int getZipFileList(QString zipFileName, QStringList *fileList) {
    unsigned long i;
    unz_global_info gi;
    int err;

    unzFile uf = openUnzipFile(zipFileName);

    if (uf == NULL) return UNZ_ERRNO;

    err = unzGetGlobalInfo(uf,&gi);

    if (err != UNZ_OK) {
        qWarning() << "Error in unzGetGlobalInfo. Error code was " << err;
        unzClose(uf);
        return err;
    }

    for (i = 0; i < gi.number_entry; i++) {
        char filename[MAXFILENAME];
        unz_file_info fileInfo;

        err = unzGetCurrentFileInfo(uf,
                                    &fileInfo,
                                    filename,
                                    sizeof(filename),
                                    NULL,
                                    0,
                                    NULL,
                                    0);
        if (err != UNZ_OK) {
            qWarning() << "Error in unzGetCurrentFileInfo. Error code was " << err;
            unzClose(uf);
            return err;
        }

        fileList->append(filename);

        if ((i+1) < gi.number_entry) {
            err = unzGoToNextFile(uf);
            if (err != UNZ_OK) {
                qWarning() << "Error in unzGoToNextFile. Error code was " << err;
                unzClose(uf);
                return err;
            }
        }
    }

    unzClose(uf);

    return UNZ_OK;
}

ZUEXPORT int extractFileFromZipFile(QString zipFileName,
                           QString fileName,
                           QIODevice *target,
                           QProgressDialog *qpd) {

    int err = UNZ_OK;
    char filenameInZip[MAXFILENAME];
    void * buffer;
    uInt bufferSize;

    /*
     * Open the zip file
     */
    unzFile uf = openUnzipFile(zipFileName);
    if (uf == NULL) return UNZ_ERRNO;

    /*
     * Try and find the file in the archive
     */
    if (unzLocateFile(uf, fileName.toAscii().constData(),CASESENSITIVITY) != UNZ_OK) {
        qWarning() << "Unable to locate file" << fileName << "in zip file" << zipFileName;
        return 2;
    }

    /*
     * Get the files information
     */
    unz_file_info fileInfo;
    err = unzGetCurrentFileInfo(uf,
                                &fileInfo,
                                filenameInZip,
                                sizeof(filenameInZip),
                                NULL,
                                0,
                                NULL,
                                0);
    if (err != UNZ_OK) {
        qWarning() << "Error in unzGetCurrentFileInfo. Error code was" << err;
        return err;
    }

    /*
     * Allocate a buffer
     */
    bufferSize = WRITEBUFFERSIZE;
    buffer = (void*)malloc(bufferSize);
    if(buffer == NULL) {
        qWarning() << "Error allocating memory for read buffer.";
        return UNZ_INTERNALERROR;
    }

    /*
     * Open the desired file in the zip archive
     */
    err = unzOpenCurrentFile(uf);
    if (err != UNZ_OK) {
        qWarning() << "Error in unzOpenCurrentFile. Error code was" << err;
        return err;
    }

    /*
     * Prepare the progress dialog if one has been provided to us.
     */
    if (qpd != NULL) {
        qpd->setMaximum(fileInfo.uncompressed_size);
    }
    unsigned long totalBytesWritten = 0;

    /*
     * Copy its data to the target device.
     */
    do {
        err = unzReadCurrentFile(uf,buffer,bufferSize);
        if (err < 0) {
            qWarning() << "Error in unzReadCurrentFile. Error code was" << err;
            break;
        }
        if (err > 0) {
            // Note that err actually contains the number of bytes read at this point.
            if (target->write((const char *)buffer,err) < err) {
                qWarning() << "Error writing to target QIODevice.";
                err = UNZ_ERRNO;
                break;
            }
            if (qpd != NULL) {
                totalBytesWritten += err;
                qpd->setValue(totalBytesWritten);
            }
        }
    } while (err > 0);

    if (qpd != NULL) qpd->setValue(fileInfo.uncompressed_size);

    /*
     * Tidy up. Be careful not to loose any error code.
     */
    if (err == UNZ_OK) {
        err = unzCloseCurrentFile(uf);
        if (err != UNZ_OK) {
            qWarning() << "Error in unzCloseCurrentFile. Error code was" << err;
        }
    } else {
        unzCloseCurrentFile(uf);
    }

    unzClose(uf);
    free(buffer);
    return err;
}
